import logging

from jasmin_cis.data_io.Coord import Coord, CoordList
from jasmin_cis.data_io.products.AProduct import AProduct
from jasmin_cis.data_io.ungridded_data import UngriddedData, Metadata
import jasmin_cis.utils as utils
import jasmin_cis.data_io.hdf as hdf
import numpy as np

class Cloudsat_2B_CWC_RVOD(AProduct):

    def get_file_signature(self):
        return [r'.*2B.CWC.RVOD.*\.hdf']

    def _generate_time_array(self, vdata):
        import jasmin_cis.data_io.hdf_vd as hdf_vd
        import datetime as dt
        from jasmin_cis.time_util import convert_sec_since_to_std_time_array

        Cloudsat_start_time = dt.datetime(1993,1,1,0,0,0)

        arrays = []
        for i,j in zip(vdata['Profile_time'],vdata['TAI_start']):
            time = hdf_vd.get_data(i)
            start = hdf_vd.get_data(j)
            time += start
            # Do the conversion to standard time here before we expand the time array...
            time = convert_sec_since_to_std_time_array(time, Cloudsat_start_time)
            arrays.append(time)
        return utils.concatenate(arrays)

    def create_coords(self, filenames):
        from jasmin_cis.time_util import cis_standard_time_unit
        # list of coordinate variables we are interested in
        variables = [ 'Latitude','Longitude','TAI_start','Profile_time','Height']

        logging.info("Listing coordinates: " + str(variables))

        # reading the various files
        sdata, vdata = hdf.read(filenames,variables)

        # altitude coordinate
        height = sdata['Height']
        height_data = hdf.read_data(height, "SD")
        height_metadata = hdf.read_metadata(height, "SD")
        height_coord = Coord(height_data, height_metadata, "Y")

        # latitude
        lat = vdata['Latitude']
        lat_data = utils.expand_1d_to_2d_array(hdf.read_data(lat, "VD"),len(height_data[0]),axis=1)
        lat_metadata = hdf.read_metadata(lat,"VD")
        lat_metadata.shape = lat_data.shape
        lat_coord = Coord(lat_data, lat_metadata)

        # longitude
        lon = vdata['Longitude']
        lon_data = utils.expand_1d_to_2d_array(hdf.read_data(lon, "VD"),len(height_data[0]),axis=1)
        lon_metadata = hdf.read_metadata(lon, "VD")
        lon_metadata.shape = lon_data.shape
        lon_coord = Coord(lon_data, lon_metadata)

        # time coordinate
        time_data = self._generate_time_array(vdata)
        time_data = utils.expand_1d_to_2d_array(time_data,len(height_data[0]),axis=1)
        time_coord = Coord(time_data,Metadata(name='Profile_time', standard_name='time', shape=time_data.shape,
                                              units=str(cis_standard_time_unit),
                                              calendar=cis_standard_time_unit.calendar),"X")


        # create object containing list of coordinates
        coords = CoordList()
        coords.append(lat_coord)
        coords.append(lon_coord)
        coords.append(height_coord)
        coords.append(time_coord)

        return coords

    def create_data_object(self, filenames, variable):

        logging.debug("Creating data object for variable " + variable)

        # reading coordinates
        coords = self.create_coords(filenames)

        # reading of variables
        sdata, vdata = hdf.read(filenames, variable)

        #missing values
        missing_values = [0,-9999,-4444,-3333]

        # retrieve data + its metadata
        if variable in vdata:
            var = hdf.read_data(vdata[variable], "VD",missing_values)
            metadata = hdf.read_metadata(vdata[variable],"VD")
        elif variable in sdata:
            var = hdf.read_data(sdata[variable], "SD",missing_values)
            metadata = hdf.read_metadata(sdata[variable],"SD")
        else:
            raise ValueError("variable not found")

        return UngriddedData(var,metadata,coords)


class MODIS_L3(AProduct):

    def _parse_datetime(self,metadata_dict,keyword):
        import re
        res = ""
        for s in metadata_dict.itervalues():
            i_start = s.find(keyword)
            ssub = s[i_start:len(s)]
            i_end = ssub.find("END_OBJECT")
            ssubsub = s[i_start:i_start+i_end]
            matches = re.findall('".*"',ssubsub)
            if len(matches) > 0:
                res = matches[0].replace('\"','')
                if res is not "":
                    break
        return res

    def _get_start_date(self, filename):
        from jasmin_cis.time_util import parse_datetimestr_to_std_time
        metadata_dict = hdf.get_hdf4_file_metadata(filename)
        date = self._parse_datetime(metadata_dict,'RANGEBEGINNINGDATE')
        time = self._parse_datetime(metadata_dict,'RANGEBEGINNINGTIME')
        datetime_str = date + " " + time
        return parse_datetimestr_to_std_time(datetime_str)

    def _get_end_date(self, filename):
        from jasmin_cis.time_util import parse_datetimestr_to_std_time
        metadata_dict = hdf.get_hdf4_file_metadata(filename)
        date = self._parse_datetime(metadata_dict,'RANGEENDINGDATE')
        time = self._parse_datetime(metadata_dict,'RANGEENDINGTIME')
        datetime_str = date + " " + time
        return parse_datetimestr_to_std_time(datetime_str)

    def get_file_signature(self):
        product_names = ['MYD08_D3','MOD08_D3',"MOD08_E3"]
        regex_list = [ r'.*' + product + '.*\.hdf' for product in product_names]
        return regex_list

    def create_coords(self, filenames):
        import numpy as np
        from jasmin_cis.time_util import calculate_mid_time, cis_standard_time_unit

        variables = ['XDim','YDim']
        logging.info("Listing coordinates: " + str(variables))

        sdata, vdata = hdf.read(filenames,variables)

        lat = sdata['YDim']
        lat_metadata = hdf.read_metadata(lat, "SD")

        lon = sdata['XDim']
        lon_metadata = hdf.read_metadata(lon, "SD")

        # expand lat and lon data array so that they have the same shape
        lat_data = utils.expand_1d_to_2d_array(hdf.read_data(lat,"SD"),lon_metadata.shape,axis=1) # expand latitude column wise
        lon_data = utils.expand_1d_to_2d_array(hdf.read_data(lon,"SD"),lat_metadata.shape,axis=0) # expand longitude row wise

        lat_metadata.shape = lat_data.shape
        lon_metadata.shape = lon_data.shape

        # to make sure "Latitude" and "Longitude", i.e. the standard_name is displayed instead of "YDim"and "XDim"
        lat_metadata.standard_name = "latitude"
        lat_metadata._name = ""
        lon_metadata.standard_name = "longitude"
        lon_metadata._name = ""

        # create arrays for time coordinate using the midpoint of the time delta between the start date and the end date
        time_data_array = []
        for filename in filenames:
            mid_datetime = calculate_mid_time(self._get_start_date(filename),self._get_end_date(filename))
            logging.debug("Using "+str(mid_datetime)+" as datetime for file "+str(filename))
            # Only use part of the full lat shape as it has already been concattenated
            time_data = np.empty((lat_metadata.shape[0]/len(filenames),lat_metadata.shape[1]),dtype='float64')
            time_data.fill(mid_datetime)
            time_data_array.append(time_data)
        time_data = utils.concatenate(time_data_array)
        time_metadata = Metadata(name='Date Time', standard_name='time', shape=time_data.shape,
                                 units=str(cis_standard_time_unit),calendar=cis_standard_time_unit.calendar)

        coords = CoordList()
        coords.append(Coord(lon_data, lon_metadata,'X'))
        coords.append(Coord(lat_data, lat_metadata,'Y'))
        coords.append(Coord(time_data, time_metadata,'T'))

        return coords


    def create_data_object(self, filenames, variable):

        logging.debug("Creating data object for variable " + variable)

        # reading coordinates
        # the variable here is needed to work out whether to apply interpolation to the lat/lon data or not
        coords = self.create_coords(filenames)

        # reading of variables
        sdata, vdata = hdf.read(filenames, variable)

        # retrieve data + its metadata
        var = sdata[variable]
        metadata = hdf.read_metadata(var, "SD")

        data = UngriddedData(var, metadata, coords)
        data.coords_on_grid = True
        return data

class MODIS_L2(AProduct):

    modis_scaling = ["1km","5km","10km"]

    def get_file_signature(self):
        product_names = ['MYD06_L2','MOD06_L2','MYD04_L2','MOD04_L2']
        regex_list = [ r'.*' + product + '.*\.hdf' for product in product_names]
        return regex_list

    def __get_data_scale(self, filename, variable):
        from jasmin_cis.exceptions import InvalidVariableError
        from pyhdf import SD

        try:
            meta = SD.SD(filename).datasets()[variable][0][0]
        except KeyError:
            raise InvalidVariableError("Variable "+variable+" not found")

        for scaling in self.modis_scaling:
            if scaling in meta:
                return scaling
        return None

    def __field_interpolate(self,data,factor=5):
        '''
        Interpolates the given 2D field by the factor,
        edge pixels are defined by the ones in the centre,
        odd factords only!
        '''
        import numpy as np

        logging.debug("Performing interpolation...")

        output = np.zeros((factor*data.shape[0],factor*data.shape[1]))*np.nan
        output[int(factor/2)::factor,int(factor/2)::factor] = data
        for i in range(1,factor+1):
            output[(int(factor/2)+i):(-1*factor/2+1):factor,:] = i*((output[int(factor/2)+factor::factor,:]-output[int(factor/2):(-1*factor):factor,:])
                                                                    /float(factor))+output[int(factor/2):(-1*factor):factor,:]
        for i in range(1,factor+1):
            output[:,(int(factor/2)+i):(-1*factor/2+1):factor] = i*((output[:,int(factor/2)+factor::factor]-output[:,int(factor/2):(-1*factor):factor])
                                                                    /float(factor))+output[:,int(factor/2):(-1*factor):factor]
        return output

    def create_coords(self, filenames, variable=None):
        import datetime as dt

        variables = [ 'Latitude','Longitude','Scan_Start_Time']
        logging.info("Listing coordinates: " + str(variables))

        sdata, vdata = hdf.read(filenames,variables)

        apply_interpolation = False
        if variable is not None:
            scale = self.__get_data_scale(filenames[0], variable)
            apply_interpolation = True if scale is "1km" else False

        lat = sdata['Latitude']
        lat_data = self.__field_interpolate(hdf.read_data(lat,"SD")) if apply_interpolation else hdf.read_data(lat,"SD")
        lat_metadata = hdf.read_metadata(lat, "SD")
        lat_coord = Coord(lat_data, lat_metadata,'Y')

        lon = sdata['Longitude']
        lon_data = self.__field_interpolate(hdf.read_data(lon,"SD")) if apply_interpolation else hdf.read_data(lon,"SD")
        lon_metadata = hdf.read_metadata(lon,"SD")
        lon_coord = Coord(lon_data, lon_metadata,'X')

        time = sdata['Scan_Start_Time']
        time_metadata = hdf.read_metadata(time,"SD")
        # Ensure the standard name is set
        time_metadata.standard_name = 'time'
        time_coord = Coord(time,time_metadata,"T")
        time_coord.convert_TAI_time_to_std_time(dt.datetime(1993,1,1,0,0,0))

        return CoordList([lat_coord,lon_coord,time_coord])

    def create_data_object(self, filenames, variable):
        logging.debug("Creating data object for variable " + variable)

        # reading coordinates
        # the variable here is needed to work out whether to apply interpolation to the lat/lon data or not
        coords = self.create_coords(filenames, variable)

        # reading of variables
        sdata, vdata = hdf.read(filenames, variable)

        # retrieve data + its metadata
        var = sdata[variable]
        metadata = hdf.read_metadata(var, "SD")

        return UngriddedData(var, metadata, coords)

class Cloud_CCI(AProduct):

    def get_file_signature(self):
        return [r'..*ESACCI.*CLOUD.*']

    def create_coords(self, filenames):

        from jasmin_cis.data_io.netcdf import read_many_files_individually, get_metadata
        from jasmin_cis.data_io.Coord import Coord

        variables = ["lat", "lon", "time"]
        logging.info("Listing coordinates: " + str(variables))

        var_data = read_many_files_individually(filenames, variables)

        coords = CoordList()
        coords.append(Coord(var_data['lat'], get_metadata(var_data['lat'][0]), 'Y'))
        coords.append(Coord(var_data['lon'], get_metadata(var_data['lon'][0]),'X'))
        time_coord = Coord(var_data['time'], get_metadata(var_data['time'][0]))

        # TODO: Is this really julian?
        time_coord.convert_julian_to_std_time()
        coords.append(time_coord)

        return coords

    def create_data_object(self, filenames, variable):

        from jasmin_cis.data_io.netcdf import get_metadata, read_many_files_individually

        coords = self.create_coords(filenames)
        var = read_many_files_individually(filenames, [variable])
        metadata = get_metadata(var[variable][0])

        return UngriddedData(var[variable], metadata, coords)

class Aerosol_CCI(AProduct):

    def get_file_signature(self):
        return [r'.*ESACCI.*AEROSOL.*']

    def create_coords(self, filenames):

        from jasmin_cis.data_io.netcdf import read_many_files, get_metadata
        from jasmin_cis.data_io.Coord import Coord
        import datetime

        variables = ["lat", "lon", "time"]
        logging.info("Listing coordinates: " + str(variables))

        data = read_many_files(filenames, variables, dim="pixel_number")

        coords = CoordList()
        coords.append(Coord(data["lon"], get_metadata(data["lon"]), "X"))
        coords.append(Coord(data["lat"], get_metadata(data["lat"]), "Y"))
        time_coord = Coord(data["time"], get_metadata(data["time"]), "T")
        time_coord.convert_TAI_time_to_std_time(datetime.datetime(1970,1,1))
        coords.append(time_coord)
        
        return coords

    def create_data_object(self, filenames, variable):
        from jasmin_cis.data_io.netcdf import read_many_files, get_metadata

        coords = self.create_coords(filenames)
        data = read_many_files(filenames, variable, dim="pixel_number")
        metadata = get_metadata(data[variable])

        return UngriddedData(data[variable], metadata, coords)


class Caliop(AProduct):

    def get_file_signature(self):
        '''
        To be implemented by subclcass
        @return:
        '''
        return []

    def create_coords(self, filenames):
        '''
        To be implemented by subclass
        @param filenames:
        @return:
        '''
        return None

    def create_data_object(self, filenames, variable):
        '''
        To be implemented by subclass
        @param filenames:
        @param variable:
        @return:
        '''
        return None

    def get_calipso_data(self, sds):
        """
        Reads raw data from an SD instance. Automatically applies the
        scaling factors and offsets to the data arrays found in Calipso data.

        Returns:
            A numpy array containing the raw data with missing data is replaced by NaN.

        Arguments:
            sds        -- The specific sds instance to read

        """
        from jasmin_cis.utils import create_masked_array_for_missing_data

        calipso_fill_values = {'Float_32' : -9999.0,
                               #'Int_8' : 'See SDS description',
                               'Int_16' : -9999,
                               'Int_32' : -9999,
                               'UInt_8' : -127,
                               #'UInt_16' : 'See SDS description',
                               #'UInt_32' : 'See SDS description',
                               'ExtinctionQC Fill Value' : 32768,
                               'FeatureFinderQC No Features Found' : 32767,
                               'FeatureFinderQC Fill Value' : 65535}

        data = sds.get()
        attributes = sds.attributes()

        # Missing data.
        missing_val = attributes.get('fillvalue', None)
        if missing_val is None:
            try:
                missing_val = calipso_fill_values[attributes.get('format', None)]
            except KeyError:
                # Last guess
                missing_val = attributes.get('_FillValue', None)

        data = create_masked_array_for_missing_data(data, missing_val)

        # Offsets and scaling.
        offset  = attributes.get('add_offset', 0)
        scale_factor = attributes.get('scale_factor', 1)
        data = self.apply_scaling_factor_CALIPSO(data, scale_factor, offset)

        return data

    def apply_scaling_factor_CALIPSO(self,data, scale_factor, offset):
        '''
        Apply scaling factor Calipso data
        @param data:
        @param scale_factor:
        @param offset:
        @return:
        '''
        return (data/scale_factor) + offset

class Caliop_L2(Caliop):

    def get_file_signature(self):
        return [r'CAL_LID_L2_05kmAPro-Prov-V3-01.*hdf']

    def create_coords(self, filenames):

        from jasmin_cis.data_io.hdf_vd import get_data
        from jasmin_cis.data_io.hdf_vd import VDS
        from pyhdf.error import HDF4Error
        from jasmin_cis.data_io import hdf_sd
        import datetime as dt
        from jasmin_cis.time_util import convert_sec_since_to_std_time_array, cis_standard_time_unit

        variables = [ 'Latitude','Longitude', "Profile_Time"]
        logging.info("Listing coordinates: " + str(variables))

        # reading data from files
        sdata = {}
        for filename in filenames:
            try:
                sds_dict = hdf_sd.read(filename, variables)
            except HDF4Error as e:
                raise IOError(str(e))

            for var in sds_dict.keys():
                utils.add_element_to_list_in_dict(sdata, var, sds_dict[var])

        alt_name = "altitude"
        logging.info("Additional coordinates: '" + alt_name + "'")

        # work out size of data arrays
        # the coordinate variables will be reshaped to match that.
        # NOTE: This assumes that all Caliop_L1 files have the same altitudes.
        #       If this is not the case, then the following line will need to be changed
        #       to concatenate the data from all the files and not just arbitrarily pick
        #       the altitudes from the first file.
        alt_data = get_data(VDS(filenames[0],"Lidar_Data_Altitudes"), True)
        len_x = alt_data.shape[0]

        lat_data = hdf.read_data(sdata['Latitude'],"SD")
        len_y = lat_data.shape[0]

        new_shape = (len_x, len_y)

        # altitude
        alt_data = utils.expand_1d_to_2d_array(alt_data,len_y,axis=0)
        alt_metadata = Metadata(name=alt_name, standard_name=alt_name, shape=new_shape)
        alt_coord = Coord(alt_data,alt_metadata)

        # latitude
        lat_data = utils.expand_1d_to_2d_array(lat_data[:,1],len_x,axis=1)
        lat_metadata = hdf.read_metadata(sdata['Latitude'], "SD")
        lat_metadata.shape = new_shape
        lat_coord = Coord(lat_data, lat_metadata, 'Y')

        # longitude
        lon = sdata['Longitude']
        lon_data = hdf.read_data(lon,"SD")
        lon_data = utils.expand_1d_to_2d_array(lon_data[:,1],len_x,axis=1)
        lon_metadata = hdf.read_metadata(lon,"SD")
        lon_metadata.shape = new_shape
        lon_coord = Coord(lon_data, lon_metadata,'X')

        #profile time, x
        time = sdata['Profile_Time']
        time_data = hdf.read_data(time,"SD")
        time_data = convert_sec_since_to_std_time_array(time_data, dt.datetime(1993,1,1,0,0,0))
        time_data = utils.expand_1d_to_2d_array(time_data[:,1],len_x,axis=1)
        time_coord = Coord(time_data,Metadata(name='Profile_Time', standard_name='time', shape=time_data.shape,
                                              units=str(cis_standard_time_unit),
                                              calendar=cis_standard_time_unit.calendar),"T")

        # create the object containing all coordinates
        coords = CoordList()
        coords.append(lat_coord)
        coords.append(lon_coord)
        coords.append(time_coord)
        coords.append(alt_coord)

        return coords


    def create_data_object(self, filenames, variable):
        logging.debug("Creating data object for variable " + variable)

        # reading coordinates
        # the variable here is needed to work out whether to apply interpolation to the lat/lon data or not
        coords = self.create_coords(filenames)

        # reading of variables
        sdata, vdata = hdf.read(filenames, variable)

        # retrieve data + its metadata
        var = sdata[variable]
        metadata = hdf.read_metadata(var, "SD")

        return UngriddedData(var, metadata, coords, self.get_calipso_data)


class Caliop_L1(Caliop):

    def get_file_signature(self):
        return [r'CAL_LID_L1-ValStage1-V3-01.*hdf']

    def create_coords(self, filenames):

        from jasmin_cis.data_io.hdf_vd import get_data
        from jasmin_cis.data_io.hdf_vd import VDS
        from pyhdf.error import HDF4Error
        from jasmin_cis.data_io import hdf_sd
        import datetime as dt
        from jasmin_cis.time_util import convert_sec_since_to_std_time_array, cis_standard_time_unit

        variables = [ 'Latitude','Longitude', "Profile_Time"]
        logging.info("Listing coordinates: " + str(variables))

        # reading data from files
        sdata = {}
        for filename in filenames:
            try:
                sds_dict = hdf_sd.read(filename, variables)
            except HDF4Error as e:
                raise IOError(str(e))

            for var in sds_dict.keys():
                utils.add_element_to_list_in_dict(sdata, var, sds_dict[var])

        alt_name = "altitude";
        logging.info("Additional coordinates: '" + alt_name + "'")

        # work out size of data arrays
        # the coordinate variables will be reshaped to match that.
        # NOTE: This assumes that all Caliop_L1 files have the same altitudes.
        #       If this is not the case, then the following line will need to be changed
        #       to concatenate the data from all the files and not just arbitrarily pick
        #       the altitudes from the first file.
        alt_data = get_data(VDS(filenames[0],"Lidar_Data_Altitudes"), True)
        len_x = alt_data.shape[0]

        lat_data = hdf.read_data(sdata['Latitude'],"SD")
        len_y = lat_data.shape[0]

        new_shape = (len_x, len_y)

        # altitude
        alt_data = utils.expand_1d_to_2d_array(alt_data,len_y,axis=0)
        alt_metadata = Metadata(name=alt_name, standard_name=alt_name, shape=new_shape)
        alt_coord = Coord(alt_data,alt_metadata)

        # latitude
        lat_data = utils.expand_1d_to_2d_array(lat_data[:,0],len_x,axis=1)
        lat_metadata = hdf.read_metadata(sdata['Latitude'], "SD")
        lat_metadata.shape = new_shape
        lat_coord = Coord(lat_data, lat_metadata, 'Y')

        # longitude
        lon = sdata['Longitude']
        lon_data = hdf.read_data(lon,"SD")
        lon_data = utils.expand_1d_to_2d_array(lon_data[:,0],len_x,axis=1)
        lon_metadata = hdf.read_metadata(lon,"SD")
        lon_metadata.shape = new_shape
        lon_coord = Coord(lon_data, lon_metadata,'X')

        #profile time, x
        time = sdata['Profile_Time']
        time_data = hdf.read_data(time,"SD")
        time_data = convert_sec_since_to_std_time_array(time_data, dt.datetime(1993,1,1,0,0,0))
        time_data = utils.expand_1d_to_2d_array(time_data[:,0],len_x,axis=1)
        time_coord = Coord(time_data,Metadata(name='Profile_Time', standard_name='time', shape=time_data.shape,
                                              units=str(cis_standard_time_unit),
                                              calendar=cis_standard_time_unit.calendar),"T")

        # create the object containing all coordinates
        coords = CoordList()
        coords.append(lat_coord)
        coords.append(lon_coord)
        coords.append(time_coord)
        coords.append(alt_coord)

        return coords


    def create_data_object(self, filenames, variable):
        logging.debug("Creating data object for variable " + variable)

        # reading coordinates
        # the variable here is needed to work out whether to apply interpolation to the lat/lon data or not
        coords = self.create_coords(filenames)

        # reading of variables
        sdata, vdata = hdf.read(filenames, variable)

        # retrieve data + its metadata
        var = sdata[variable]
        metadata = hdf.read_metadata(var, "SD")

        return UngriddedData(var, metadata, coords, self.get_calipso_data)


class Cis(AProduct):

    def get_file_signature(self):
        return [r'cis\-.*\.nc']

    def create_coords(self, filenames, usr_variable = None):
        from jasmin_cis.data_io.netcdf import read_many_files_individually, get_metadata
        from jasmin_cis.data_io.Coord import Coord
        from jasmin_cis.exceptions import InvalidVariableError

        variables = [ ("longitude", "x"), ("latitude","y"), ("altitude","z"), ("time","t") ]

        logging.info("Listing coordinates: " + str(variables))

        coords = CoordList()
        for variable in variables:
            try:
                var_data = read_many_files_individually(filenames,variable[0])[variable[0]]
                coords.append(Coord(var_data, get_metadata(var_data[0]),axis=variable[1]))
            except InvalidVariableError:
                pass

        # Note - We don't need to convert this time coord as it should have been written in our
        #  'standard' time unit

        if usr_variable is None:
            res = coords
        else:
            usr_var_data = read_many_files_individually(filenames,usr_variable)[usr_variable]
            res = UngriddedData(usr_var_data, get_metadata(usr_var_data[0]), coords)

        return res

    def create_data_object(self, filenames, variable):
        return self.create_coords(filenames, variable)

class NetCDF_CF(AProduct):

    def get_file_signature(self):
        # We don't know of any 'standard' netCDF CF data yet...
        return []

    def create_coords(self, filenames, variable = None):
        from jasmin_cis.data_io.netcdf import read_many_files, get_metadata
        from jasmin_cis.data_io.Coord import Coord

        variables = [ "latitude", "longitude", "altitude", "time" ]
        logging.info("Listing coordinates: " + str(variables))

        if variable is not None:
            variables.append(variable)

        data_variables = read_many_files(filenames, variables)

        coords = CoordList()
        coords.append(Coord(data_variables["longitude"], get_metadata(data_variables["longitude"]), "X"))
        coords.append(Coord(data_variables["latitude"], get_metadata(data_variables["latitude"]), "Y"))
        coords.append(Coord(data_variables["altitude"], get_metadata(data_variables["altitude"]), "Z"))
        coords.append(Coord(data_variables["time"], get_metadata(data_variables["time"]), "T"))

        if variable is None:
            return coords
        else:
            return UngriddedData(data_variables[variable], get_metadata(data_variables[variable]), coords)

    def create_data_object(self, filenames, variable):
        return self.create_coords(filenames, variable)


class NCAR_NetCDF_RAF(NetCDF_CF):

    def get_file_signature(self):
        return [r'RF.*\.nc']

    def create_coords(self, filenames, variable = None):
        from jasmin_cis.data_io.netcdf import read_many_files, get_metadata
        from jasmin_cis.data_io.Coord import Coord

        variables = [ "LATC", "LONC", "GGALTC", "Time" ]
        logging.info("Listing coordinates: " + str(variables))

        if variable is not None:
            variables.append(variable)

        data_variables = read_many_files(filenames, variables, dim='Time')

        coords = CoordList()
        coords.append(Coord(data_variables["LATC"], get_metadata(data_variables["LATC"]), "Y"))
        coords.append(Coord(data_variables["LONC"], get_metadata(data_variables["LONC"]), "X"))
        coords.append(Coord(data_variables["GGALTC"], get_metadata(data_variables["GGALTC"]), "Z"))
        time_coord = Coord(data_variables["Time"], get_metadata(data_variables["Time"]), "T")
        time_coord.convert_to_std_time()
        coords.append(time_coord)

        if variable is None:
            return coords
        else:
            return UngriddedData(data_variables[variable], get_metadata(data_variables[variable]), coords)

    def create_data_object(self, filenames, variable):
        return self.create_coords(filenames, variable)


class NetCDF_CF_Gridded(NetCDF_CF):

    def get_file_signature(self):
        # We don't know of any 'standard' netCDF CF model data yet...
        return []

    def create_coords(self, filenames, variable = None):
        # TODO Expand coordinates
        # For gridded data sets this will actually return coordinates which are too short
        #  we need to think about how to expand them here
        """

        @param filenames: List of filenames to read coordinates from
        @param variable: Optional variable to read while we're reading the coordinates
        @return: If variable was specified this will return an UngriddedData object, otherwise a CoordList
        """
        return super(NetCDF_CF_Gridded, self).create_coords(filenames, variable)

    def create_data_object(self, filenames, variable):
        """

        @param filenames: List of filenames to read coordinates from
        @param variable: Optional variable to read while we're reading the coordinates, can be a string or a VariableConstraint object
        @return: If variable was specified this will return an UngriddedData object, otherwise a CoordList
        """
        from jasmin_cis.exceptions import InvalidVariableError
        import iris

        # checking if the files given actually exist
        for filename in filenames:
            with open(filename) as f: pass

        try:
            cube = iris.load_cube(filenames, variable)
        except iris.exceptions.ConstraintMismatchError:
            raise InvalidVariableError("Variable not found: " + str(variable) +
                                       "\nTo see a list of variables run: cis info " + filenames[0] + " -h")
        except ValueError as e:
            raise IOError(e)

        sub_cube = list(cube.slices([ coord for coord in cube.coords(dim_coords=True) if coord.points.size > 1]))[0]
        #  Ensure that there are no extra dimensions which can confuse the plotting.
        # E.g. the shape of the cube might be (1, 145, 165) and so we don't need to know about
        #  the dimension whose length is one. The above list comprehension would return a cube of
        #  shape (145, 165)

        return sub_cube

class Xglnwa_vprof(NetCDF_CF_Gridded):

    def get_file_signature(self):
        return [r'.*xglnwa.*vprof.*\.nc']

    def create_coords(self, filenames, variable = None):
        from jasmin_cis.data_io.netcdf import read_many_files, get_metadata
        from jasmin_cis.data_io.Coord import Coord

        variables = [ "latitude" ]
        logging.info("Listing coordinates: " + str(variables))

        if variable is not None:
            variables.append(variable)

        data_variables = read_many_files(filenames, variables)

        coords = CoordList()
        coords.append(Coord(data_variables["latitude"], get_metadata(data_variables["latitude"]), "X"))

        if variable is None:
            return coords
        else:
            return UngriddedData(data_variables[variable], get_metadata(data_variables[variable]), coords)


    def create_data_object(self, filenames, variable):
        from iris import AttributeConstraint
        # In this case we use the variable as a name constraint as the variable names themselves aren't obvious
        var_constraint = AttributeConstraint(name=variable)

        return super(Xglnwa_vprof, self).create_data_object(filenames, var_constraint)

class Xglnwa(NetCDF_CF_Gridded):

    def get_file_signature(self):
        return [r'.*xglnwa.*\.nc']

    def create_coords(self, filenames, variable = None):
        # TODO Expand coordinates
        # For gridded data sets this will actually return coordinates which are too short
        #  we need to think about how to expand them here
        """

        @param filenames: List of filenames to read coordinates from
        @param variable: Optional variable to read while we're reading the coordinates
        @return: If variable was specified this will return an UngriddedData object, otherwise a CoordList
        """
        return super(Xglnwa, self).create_coords(filenames, variable)


    def create_data_object(self, filenames, variable):
        from iris import AttributeConstraint
        # In this case we use the variable as a name constraint as the variable names themselves aren't obvious
        var_constraint = AttributeConstraint(name=variable)

        return super(Xglnwa, self).create_data_object(filenames, var_constraint)

class Xenida(NetCDF_CF_Gridded):

    def get_file_signature(self):
        return [r'.*xenida.*\.nc']

    def create_coords(self, filenames, variable=None):
        # TODO Expand coordinates and read multiple files
        # For gridded data sets this will actually return coordinates which are too short
        #  we need to think about how to expand them here
        """

        @param filenames: List of filenames to read coordinates from
        @param variable: Optional variable to read while we're reading the coordinates
        @return: If variable was specified this will return an UngriddedData object, otherwise a CoordList
        """
        from jasmin_cis.data_io.netcdf import read, get_metadata
        from jasmin_cis.data_io.Coord import Coord

        variables = [ "latitude", "longitude", "atmosphere_hybrid_height_coordinate_ak", "time" ]
        logging.info("Listing coordinates: " + str(variables))

        data_variables = read(filenames[0], variables)

        coords = CoordList()
        coords.append(Coord(data_variables["longitude"], get_metadata(data_variables["longitude"]), "X"))
        coords.append(Coord(data_variables["latitude"], get_metadata(data_variables["latitude"]), "Y"))


        altitude = Coord(data_variables["atmosphere_hybrid_height_coordinate_ak"], get_metadata(data_variables["atmosphere_hybrid_height_coordinate_ak"]), "Z")
        altitude.standard_name = 'altitude'


        coords.append(altitude)

        #days since 1979-4-1
        time = Coord(data_variables["time"], get_metadata(data_variables["time"]), "T")
        time.convert_to_std_time()
        coords.append(time)

        return coords

    def create_data_object(self, filenames, variable):
        from iris.aux_factory import HybridHeightFactory
        from jasmin_cis.time_util import convert_cube_time_coord_to_standard_time

        cube = super(Xenida, self).create_data_object(filenames, variable)
        cube.add_aux_factory(HybridHeightFactory(cube.coords()[5]))
        cube = convert_cube_time_coord_to_standard_time(cube)

        return cube

class Aeronet(AProduct):

    def get_file_signature(self):
        return [r'.*\.lev20']

    def create_coords(self, filenames, data = None):
        from jasmin_cis.data_io.ungridded_data import Metadata
        from jasmin_cis.data_io.aeronet import load_multiple_aeronet

        if data is None:
            data = load_multiple_aeronet(filenames)

        coords = CoordList()
        coords.append(Coord(data['longitude'], Metadata(name="Longitude", shape=(len(data),), units="degrees_east", range=(-180,180))))
        coords.append(Coord(data['latitude'], Metadata(name="Latitude", shape=(len(data),), units="degrees_north", range=(-90,90))))
        coords.append(Coord(data['altitude'], Metadata(name="Altitude", shape=(len(data),), units="meters", range=(-90,90))))
        time_coord = Coord(data["datetime"], Metadata(name="Date time",standard_name='time', shape=(len(data),), units="DateTime Object"), "X")
        time_coord.convert_datetime_to_standard_time()
        coords.append(time_coord)

        return coords

    def create_data_object(self, filenames, variable):
        from jasmin_cis.data_io.aeronet import load_multiple_aeronet
        from jasmin_cis.exceptions import InvalidVariableError

        try:
            data_obj = load_multiple_aeronet(filenames, [variable])
        except ValueError:
            raise InvalidVariableError(variable + " does not exist in " + str(filenames))

        coords = self.create_coords(filenames, data_obj)

        return UngriddedData(data_obj[variable], Metadata(name=variable, long_name=variable,shape=(len(data_obj),)), coords)

class ASCII_Hyperpoints(AProduct):

    def get_file_signature(self):
        return [r'.*\.txt']

    def create_coords(self, filenames, variable=None):
        from jasmin_cis.data_io.ungridded_data import Metadata
        from numpy import genfromtxt
        from jasmin_cis.exceptions import InvalidVariableError
        from jasmin_cis.time_util import parse_datetimestr_to_std_time_array

        array_list = []

        for filename in filenames:
            try:
                array_list.append(genfromtxt(filename, dtype="f8,f8,f8,S20,f8",
                                             names=['latitude', 'longitude', 'altitude', 'time', 'value'],
                                             delimiter=',', missing_values='', usemask=True, invalid_raise=True))
            except:
                raise IOError('Unable to read file '+filename)

        data_array = utils.concatenate(array_list)
        n_elements = len(data_array['latitude'])

        coords = CoordList()
        coords.append(Coord(data_array["latitude"], Metadata(standard_name="latitude", shape=(n_elements,), units="degrees_north")))
        coords.append(Coord(data_array["longitude"], Metadata(standard_name="longitude", shape=(n_elements,), units="degrees_east")))
        coords.append(Coord(data_array["altitude"], Metadata(standard_name="altitude", shape=(n_elements,), units="meters")))

        time_arr = parse_datetimestr_to_std_time_array(data_array["time"])
        time = Coord(time_arr, Metadata(standard_name="time", shape=(n_elements,), units="DateTime Object"))
        coords.append(time)

        if variable:
            try:
                data = UngriddedData(data_array['value'], Metadata(name="Altitude", shape=(n_elements,), units="meters"), coords)
            except:
                InvalidVariableError("Value column does not exist in file " + filenames)
            return data
        else:
            return coords

    def create_data_object(self, filenames, variable):
        return self.create_coords(filenames, True)



import matplotlib.pyplot as plt
from mpl_toolkits.basemap import Basemap
import logging

class Generic_Plot(object):
    DEFAULT_NUMBER_OF_COLOUR_BAR_STEPS = 5

    def __init__(self, packed_data_items, plot_args, calculate_min_and_max_values = True, *mplargs, **mplkwargs):
        '''
        Constructor for Generic_Plot.
        Note: This also calls the plot method
        @param packed_data_items: A list of packed (i.e. Iris cubes or Ungridded data objects) data items
        @param plot_args: A dictionary of plot args that was created by plot.py
        @param mplargs: Any arguments to be passed directly into matplotlib
        @param mplkwargs: Any keyword arguments to be passed directly into matplotlib
        '''
        self.mplargs = mplargs
        self.mplkwargs = mplkwargs

        if plot_args.get("logv", False):
            from matplotlib.colors import LogNorm
            self.mplkwargs["norm"] = LogNorm()

        self.plot_args = plot_args
        self.packed_data_items = packed_data_items

        self.assign_variables_to_x_and_y_axis()

        logging.debug("Unpacking the data items")
        self.unpacked_data_items = self.unpack_data_items()
        if calculate_min_and_max_values: self.calculate_min_and_max_values()

        self.matplotlib = plt

        self.plotting_library = self.set_plotting_library()

        self.set_width_and_height()
        
        self.plot()

    def set_plotting_library(self):
        if self.is_map():
            self.basemap = Basemap(lon_0=(self.unpacked_data_items[0])["x"].max()-180.0)
            return self.basemap
        else:
            return self.matplotlib

    def unpack_data_items(self):
        def __get_data(axis):
            variable = self.plot_args[axis + "_variable"]
            if variable == "default" or variable == self.packed_data_items[0].standard_name or variable == self.packed_data_items[0].long_name:
                return self.packed_data_items[0].data
            else:
                if variable.startswith("search"):
                    number_of_points = float(variable.split(":")[1])
                    for coord in self.packed_data_items[0].coords():
                        if coord.shape[0] == number_of_points: break
                else:
                    coord = self.packed_data_items[0].coord(variable)
                return coord.points if isinstance(self.packed_data_items[0], Cube) else coord.data

        def __set_variable_as_data(axis):
            old_variable = self.plot_args[axis + "_variable"]
            self.plot_args[axis + "_variable"] = self.packed_data_items[0].name()
            logging.info("Plotting " + self.plot_args[axis + "_variable"] + " on the " + axis + " axis as " + old_variable + " has length 1")

        def __swap_x_and_y_variables():
            temp =  self.plot_args["x_variable"]
            self.plot_args["x_variable"] = self.plot_args["y_variable"]
            self.plot_args["y_variable"] = temp


        from jasmin_cis.utils import unpack_data_object
        from iris.cube import Cube
        import logging
        if len(self.packed_data_items[0].shape) == 1:
            x_data = __get_data("x")
            y_data = __get_data("y")

            if len(x_data) == 1 and len(y_data) == len(self.packed_data_items[0].data):
                __set_variable_as_data("x")
            elif len(y_data) == 1 and len(x_data) == len(self.packed_data_items[0].data):
                __set_variable_as_data("y")
            else:
                try:
                    if (x_data == y_data).all():
                        __swap_x_and_y_variables()
                except AttributeError:
                    if x_data == y_data:
                        __swap_x_and_y_variables()

        return [unpack_data_object(packed_data_item, self.plot_args["x_variable"], self.plot_args["y_variable"]) for packed_data_item in self.packed_data_items]

    def unpack_comparative_data(self):
        return [{"data" : packed_data_item.data} for packed_data_item in self.packed_data_items]

    def plot(self):
        '''
        The method that will do the plotting. To be implemented by each subclass of Generic_Plot.
        '''
        raise NotImplementedError()

    def format_plot(self):
        '''
        The method that will format the plot. To be implemented by each subclass of Generic_Plot.
        '''
        raise NotImplementedError()

    def set_default_axis_label(self, axis):
        '''
        The method that will set the default axis label. To be implemented by each subclass of Generic_Plot.
        @param axis: The axis of which to set the default label for. Either "x" or "y".
        '''
        raise NotImplementedError()

    def create_legend(self):
        '''
        Creates a draggable legend in the "best" location for the plot.
        Works out legend labels unless explicitly given to the parser in the datagroups argument.
        '''
        legend_titles = []
        datagroups = self.plot_args["datagroups"]
        for i, item in enumerate(self.packed_data_items):
            if datagroups is not None and datagroups[i]["label"]:
                legend_titles.append(datagroups[i]["label"])
            else:
                legend_titles.append(item.long_name)
        legend = self.matplotlib.legend(legend_titles, loc="best")
        legend.draggable(state = True)

    def calculate_axis_limits(self, axis, min_val, max_val, step):
        '''
        Calculates the axis limits for a given axis
        @param axis: The axis for the limits to be calculated
        @return: A dictionary containing the min and max values of an array along a given axis
        '''
        calculated_min, calculated_max = self.calculate_min_and_max_values_of_array_including_case_of_log(axis, self.unpacked_data_items[0][axis])
        valrange = {}
        valrange[axis + "min"] = calculated_min if min_val is None else min_val
        valrange[axis + "max"] = calculated_max if max_val is None else max_val
        valrange[axis + "step"] = step

        return valrange

    def apply_axis_limits(self, valrange, axis):
        '''
        Applies the limits to the specified axis if given, or calculates them otherwise
        @param valrange    A dictionary containing xmin, xmax or ymin, ymax
        @param axis        The axis to apply the limits to
        '''
        valrange = self.calculate_axis_limits(axis, valrange.get(axis + "min", None), valrange.get(axis + "max", None), valrange.get(axis + "step", None))

        if axis == "x":
            step = valrange.pop("xstep", None)
            self.matplotlib.xlim(**valrange)
            if step is not None: valrange["xstep"] = step
        elif axis == "y":
            step = valrange.pop("ystep", None)
            self.matplotlib.ylim(**valrange)
            if step is not None: valrange["ystep"] = step

    def add_color_bar(self):
        '''
        Adds a colour bar to a plot
        Allows specifying of tick spacing and orientation
        '''

        step = self.plot_args["valrange"].get("vstep", None)
        if step is None:
            ticks = None
        else:
            from matplotlib.ticker import MultipleLocator
            ticks = MultipleLocator(step)

        cbar = self.matplotlib.colorbar(orientation = self.plot_args["cbarorient"], ticks = ticks)

        if not self.plot_args["logv"]:
            cbar.formatter.set_scientific(True)
            cbar.formatter.set_powerlimits((-3,3))
            cbar.update_ticks()

        if self.plot_args["cbarlabel"] is None:
            label = self.format_units(self.packed_data_items[0].units)
        else:
            label = self.plot_args["cbarlabel"]

        cbar.set_label(label)

    def set_axis_ticks(self, axis, no_of_dims):
        from numpy import arange

        if axis == "x":
            coord_axis = "x"
            tick_method = self.matplotlib.xticks
        elif axis == "y":
            coord_axis = "data" if no_of_dims == 2 else "y"
            tick_method = self.matplotlib.yticks

        if self.plot_args.get(axis + "tickangle", None) is None:
            angle = None
            ha = "center" if axis == "x" else "right"
        else:
            angle = self.plot_args[axis + "tickangle"]
            ha = "right"

        if self.is_map() and self.plot_args[axis + "range"].get(axis + "step") is None:
            self.plot_args[axis + "range"][axis + "step"] = 30

        if self.plot_args[axis + "range"].get(axis + "step") is not None:
            step = self.plot_args[axis + "range"][axis + "step"]

            if self.plot_args[axis + "range"].get(axis + "min") is None:
                min_val = min(unpacked_data_item[coord_axis].min() for unpacked_data_item in self.unpacked_data_items)
            else:
                min_val = self.plot_args[axis + "range"][axis + "min"]

            if self.plot_args[axis + "range"].get(axis + "max") is None:
                max_val = max(unpacked_data_item[coord_axis].max() for unpacked_data_item in self.unpacked_data_items)
            else:
                max_val = self.plot_args[axis + "range"][axis + "max"]

            ticks = arange(min_val, max_val+step, step)

            tick_method(ticks, rotation=angle, ha=ha)
        else:
            tick_method(rotation=angle, ha=ha)

    def format_time_axis(self):
        from jasmin_cis.time_util import cis_standard_time_unit

        coords = self.packed_data_items[0].coords(name=self.plot_args["x_variable"])

        if len(coords) == 1:
            if coords[0].units == str(cis_standard_time_unit):
                self.set_x_axis_as_time()

    def set_default_axis_label_for_comparative_plot(self, axis):
        '''
        Sets the default axis label for a comparative plot, e.g. a comparative scatter or a 3d histogram
        @param axis: The axis to set the default label for
        '''
        axis = axis.lower()
        axislabel = axis + "label"
        if axis == 'x':
            item_index = 0
        elif axis == 'y':
            item_index = 1

        if self.plot_args[axislabel] is None:
            units = self.packed_data_items[item_index].units
            name = self.packed_data_items[item_index].name()
            self.plot_args[axislabel] = name + " " + self.format_units(units)

    def set_width_and_height(self):
        '''
        Sets the width and height of the plot
        Uses an aspect ratio of 4:3 if only one of width and height are specified
        '''
        height = self.mplkwargs.pop("height", None)
        width = self.mplkwargs.pop("width", None)

        if height is not None:
            if width is None:
                width = height * (4.0 / 3.0)
        elif width is not None:
            height = width * (3.0 / 4.0)

        if height is not None and width is not None:
            self.matplotlib.figure(figsize = (width, height))

    def calculate_min_and_max_values_of_array_including_case_of_log(self, axis, array):
        '''
        Calculates the min and max values of a given array.
        If a log scale is being used on the given axis, only positive values are taken into account
        @param axis: The axis to check if a log scale is being used for
        @param array: The array to calculate the min and max values of
        @return: The min and max values of the array
        '''
        log_axis = self.plot_args.get("log" + axis, False)

        if log_axis:
            import numpy.ma as ma
            positive_array = ma.array(array, mask=array<=0)
            min_val = positive_array.min()
            max_val = positive_array.max()
        else:
            min_val =  array.min()
            max_val =  array.max()
        return min_val, max_val

    def set_x_axis_as_time(self):
        from matplotlib import ticker
        from jasmin_cis.time_util import convert_std_time_to_datetime

        ax = self.matplotlib.gca()
        def format_date(x, pos=None):
            return convert_std_time_to_datetime(x).strftime('%Y-%m-%d')

        def format_datetime(x, pos=None):
            # use iosformat rather than strftime as strftime can't handle dates before 1900 - the output is the same
            date_time = convert_std_time_to_datetime(x)
            if date_time.hour == 0 and date_time.minute == 0 and date_time.second == 0:
                return date_time.strftime("%Y-%m-%d")
            elif date_time.second == 0:
                return date_time.strftime("%Y-%m-%d %H:%M")
            else:
                return date_time.isoformat(' ')

        def format_time(x, pos=None):
            return convert_std_time_to_datetime(x).strftime('%H:%M:%S')

        ax.xaxis.set_major_formatter(ticker.FuncFormatter(format_datetime))
        tick_angle = self.plot_args.get("xtickangle", None)
        if tick_angle is None:
            self.plot_args["xtickangle"] = 45
        self.matplotlib.xticks(rotation=self.plot_args["xtickangle"], ha="right")
        # Give extra spacing at bottom of plot due to rotated labels
        self.matplotlib.gcf().subplots_adjust(bottom=0.3)

        #ax.xaxis.set_minor_formatter(ticker.FuncFormatter(format_time))

    def set_font_size(self):
        '''
        Converts the fontsize argument (if specified) from a float into a dictionary that matplotlib can recognise.
        Could be further extended to allow specifying bold, and other font formatting
        '''
        if self.plot_args.get("fontsize", None) is not None:
            if not isinstance(self.plot_args.get("fontsize", None), dict):
                self.plot_args["fontsize"] = { "font.size" : float(self.plot_args["fontsize"]) }
        else:
            self.plot_args.pop("fontsize", None)

    def is_map(self):
        '''
        @return: A boolean saying if the first packed data item contains lat and lon coordinates
        '''
        from iris.exceptions import CoordinateNotFoundError as irisNotFoundError
        from jasmin_cis.exceptions import CoordinateNotFoundError as JasminNotFoundError
        try:
            x = self.packed_data_items[0].coord(name=self.plot_args["x_variable"])
            y = self.packed_data_items[0].coord(name=self.plot_args["y_variable"])
        except (JasminNotFoundError, irisNotFoundError):
            return False

        if x.name().lower().startswith("lon") and y.name().lower().startswith("lat"):
            return True
        else:
            return False

    def calculate_min_and_max_values(self):
        '''
        Calculates the min and max values of all the data given
        Stores the values in the matplotlib keyword args to be directly passed into the plot methods.
        '''
        from sys import maxint
        from numpy import ma

        if self.plot_args.get('logv', False):
            mask_val = 0.0
        else:
            mask_val = -maxint - 1

        vmin = self.plot_args["valrange"].get("vmin", maxint)
        vmax = self.plot_args["valrange"].get("vmax", -maxint - 1)

        if vmin == maxint:
            vmin = min(ma.array(unpacked_data_item["data"], mask=unpacked_data_item["data"]<=mask_val).min() for unpacked_data_item in self.unpacked_data_items)

        if vmax == -maxint - 1:
            vmax = max(ma.array(unpacked_data_item["data"], mask=unpacked_data_item["data"]<=mask_val).max() for unpacked_data_item in self.unpacked_data_items)

        self.mplkwargs["vmin"] = float(vmin)
        self.mplkwargs["vmax"] = float(vmax)

    def contour_plot(self, filled):
        '''
        Used by both contour and contourf to plot a contour plot
        @param filled: A boolean specifying whether or not the contour plot should be filled
        '''
        from numpy import linspace
        vmin = self.mplkwargs.pop("vmin")
        vmax = self.mplkwargs.pop("vmax")

        if self.plot_args["valrange"].get("vstep", None) is None:
            step = self.DEFAULT_NUMBER_OF_COLOUR_BAR_STEPS + 1
        else:
            step = (vmax - vmin) / self.plot_args["valrange"]["vstep"]

        if filled:
            contour_type = self.plotting_library.contourf
        else:
            contour_type = self.plotting_library.contour

        if self.is_map(): self.mplkwargs["latlon"] = True

        # If data (and x, y) is one dimensional, "tri" is set to true to tell Basemap the data is unstructured
        if self.unpacked_data_items[0]["data"].ndim == 1:
            self.mplkwargs["tri"] = True

        contour_type(self.unpacked_data_items[0]["x"], self.unpacked_data_items[0]["y"], self.unpacked_data_items[0]["data"], linspace(vmin, vmax, step), *self.mplargs, **self.mplkwargs)

        self.mplkwargs.pop("latlon", None)
        self.mplkwargs.pop("tri", None)
        
        self.mplkwargs["vmin"] = vmin
        self.mplkwargs["vmax"] = vmax

    def set_log_scale(self, logx, logy):
        '''
        Sets a log (base 10) scale (if specified) on the axes
        @param logx: A boolean specifying whether or not to apply a log scale to the x axis
        @param logy: A boolean specifying whether or not to apply a log scale to the y axis
        '''
        if logx: self.matplotlib.xscale("log")
        if logy: self.matplotlib.yscale("log")

    def set_axes_ticks(self, no_of_dims):
        self.set_axis_ticks("x", no_of_dims)
        self.set_axis_ticks("y", no_of_dims)

    def format_2d_plot(self):
        '''
        Used by 2d subclasses to format the plot
        '''
        import logging
        from jasmin_cis.plotting.plot import plot_options

        logx = self.plot_args.get("logx", False)
        logy = self.plot_args.get("logy", False)
        if logx or logy:
            self.set_log_scale(logx, logy)
        else:
            try:
                self.matplotlib.gca().ticklabel_format(style='sci', scilimits=(-3,3), axis='both')
            except AttributeError:
                logging.warning("Couldn't apply scientific notation to axes")

        draw_grid = self.plot_args.pop("grid", False)
        if draw_grid: self.matplotlib.grid(True, which="both")

        self.set_axes_ticks(2)

        self.set_font_size()

        # If any of the options have not been specified, then use the defaults
        self.set_default_axis_label("X")
        self.set_default_axis_label("Y")

        if self.plot_args["xlabel"] is None: self.plot_args["xlabel"] = ""
        if self.plot_args["ylabel"] is None: self.plot_args["ylabel"] = ""
        if self.plot_args["title"] is None: self.plot_args["title"] = ""

        for key in plot_options.keys():
            # Call the method associated with the option
            if key in self.plot_args.keys():
                plot_options[key](self.plot_args[key])

        if len(self.packed_data_items) > 1: self.create_legend()

    def drawcoastlines(self):
        colour = self.plot_args["coastlinescolour"] if self.plot_args["coastlinescolour"] is not None else "k"
        self.basemap.drawcoastlines(color = colour)

    def format_3d_plot(self):
        '''
        Used by 3d subclasses to format the plot
        '''
        from jasmin_cis.plotting.plot import plot_options

        logx = self.plot_args.get("logx", False)
        logy = self.plot_args.get("logy", False)
        if logx or logy:
            self.set_log_scale(logx, logy)

        draw_grid = self.plot_args.get("grid")
        if draw_grid: self.matplotlib.grid(True, which="both")

        if self.is_map(): self.drawcoastlines()

        self.set_axes_ticks(3)

        self.set_font_size()
        # If any of the options have not been specified, then use the defaults
        self.set_default_axis_label("X")
        self.set_default_axis_label("Y")

        if self.plot_args["xlabel"] is None: self.plot_args["xlabel"] = ""
        if self.plot_args["ylabel"] is None: self.plot_args["ylabel"] = ""
        if self.plot_args["title"] is None: self.plot_args["title"] = self.packed_data_items[0].long_name

        for key in plot_options.keys():
        # Call the method associated with the option
            if key in self.plot_args.keys():
                plot_options[key](self.plot_args[key])

        if not self.plot_args["nocolourbar"]: self.add_color_bar()

        if len(self.packed_data_items) > 1: self.create_legend()

    def set_3daxis_label(self, axis):
        '''
        Used by 3d plots to calculate the default axis label.
        Uses Latitude or Longitude if a map is being plotted
        @param axis: The axis to calculate the default label for
        '''
        import jasmin_cis.exceptions as cisex
        import iris.exceptions as irisex
        axis = axis.lower()
        axislabel = axis + "label"

        if self.plot_args[axislabel] is None:
            if self.is_map():
                self.plot_args[axislabel] = "Longitude" if axis == "x" else "Latitude"
            else:
                try:
                    name = self.packed_data_items[0].coord(name=self.plot_args[axis + "_variable"]).name()
                except (cisex.CoordinateNotFoundError, irisex.CoordinateNotFoundError):
                    name = self.packed_data_items[0].name()

                try:
                    units = self.packed_data_items[0].coord(name=self.plot_args[axis + "_variable"]).units
                except (cisex.CoordinateNotFoundError, irisex.CoordinateNotFoundError):
                    units = self.packed_data_items[0].units

                # in general, display both name and units in brackets
                self.plot_args[axislabel] = name + " " + self.format_units(units)

    def format_units(self, units):
        '''
        @param units: The units of a variable, as a string
        @return: The units surrounding brackets, or the empty string if no units given
        '''
        if units:
            return "(" + str(units) + ")"
        else:
            return ""

    def assign_variables_to_x_and_y_axis(self):
        '''
        Overwrites which variable to used for the x and y axis
        Does not work for Iris Cubes
        @param main_arguments: The arguments received from the parser
        @param data: A list of packed data objects
        '''
        import logging
        from jasmin_cis.exceptions import NotEnoughAxesSpecifiedError

        x_variable = self.get_variable_name("x")
        y_variable = self.get_variable_name("y")

        if x_variable == y_variable:
            specified_axis = "x" if self.plot_args["x_variable"] is not None else "y"
            not_specified_axis = "y" if specified_axis == "x" else "y"
            raise NotEnoughAxesSpecifiedError("--" + not_specified_axis + "axis must also be specified if assigning the current " + not_specified_axis + " axis coordinate to the " + specified_axis + " axis")

        if "search" in x_variable:
            logging.info("Plotting unknown on the x axis")
        else:
            logging.info("Plotting " + x_variable + " on the x axis")

        if "search" in y_variable:
            logging.info("Plotting unknown on the y axis")
        else:
            logging.info("Plotting " + y_variable + " on the y axis")

        self.plot_args["x_variable"] = x_variable
        self.plot_args["y_variable"] = y_variable

    def get_variable_name(self, axis):
        import iris.exceptions as iris_ex
        import jasmin_cis.exceptions as jasmin_ex

        # If the user has explicitly specified what variable they want plotting on the axis
        if self.plot_args[axis + '_variable'] is None:
            try:
                return self.packed_data_items[0].coord(axis=axis.upper()).name()
            except (iris_ex.CoordinateNotFoundError, jasmin_ex.CoordinateNotFoundError):
                if axis == "x":
                    number_of_points_in_dimension = self.packed_data_items[0].shape[0]
                elif axis == "y":
                    if len(self.packed_data_items[0].shape) > 1:
                        number_of_points_in_dimension = self.packed_data_items[0].shape[1]
                    else:
                        return "default"

                for coord in self.packed_data_items[0].coords():
                    if coord.shape[0] == number_of_points_in_dimension:
                        return "search:" + str(number_of_points_in_dimension)
                return "default"
        else:
            return self.plot_args[axis + "_variable"]



from jasmin_cis.col_framework import Colocator, Constraint, Kernel
import logging


class DefaultColocator(Colocator):

    def __init__(self, var_name='', var_long_name='', var_units=''):
        super(DefaultColocator, self).__init__()
        self.var_name = var_name
        self.var_long_name = var_long_name
        self.var_units = var_units

    def colocate(self, points, data, constraint, kernel):
        '''
            This colocator takes a list of HyperPoints and a data object (currently either Ungridded data or a Cube) and returns
             one new LazyData object with the values as determined by the constraint and kernel objects. The metadata
             for the output LazyData object is copied from the input data object.
        @param points: A list of HyperPoints
        @param data: An UngriddedData object or Cube, or any other object containing metadata that the constraint object can read
        @param constraint: An instance of a Constraint subclass which takes a data object and returns a subset of that data
                            based on it's internal parameters
        @param kernel: An instance of a Kernel subclass which takes a numberof points and returns a single value
        @return: A single LazyData object
        '''
        from jasmin_cis.data_io.ungridded_data import LazyData, UngriddedData
        import numpy as np

        metadata = data.metadata

        # Convert ungridded data to a list of points
        if isinstance(data, UngriddedData):
            data = data.get_non_masked_points()

        logging.info("--> colocating...")

        # Fill will the FillValue from the start
        values = np.zeros(len(points)) + constraint.fill_value

        for i, point in enumerate(points):
            con_points = constraint.constrain_points(point, data)
            try:
                values[i] = kernel.get_value(point, con_points)
            except ValueError:
                pass
        new_data = LazyData(values, metadata)
        if self.var_name: new_data.metadata._name = self.var_name
        if self.var_long_name: new_data.metadata.long_name = self.var_long_name
        if self.var_units: new_data.units = self.var_units
        new_data.metadata.shape = (len(points),)
        new_data.metadata.missing_value = constraint.fill_value

        return [new_data]


class AverageColocator(Colocator):

    def __init__(self, var_name='', var_long_name='', var_units='',stddev_name='',nopoints_name=''):
        super(AverageColocator, self).__init__()
        self.var_name = var_name
        self.var_long_name = var_long_name
        self.var_units = var_units
        self.stddev_name = stddev_name
        self.nopoints_name = nopoints_name

    def colocate(self, points, data, constraint, kernel):
        '''
            This colocator takes a list of HyperPoints and a data object (currently either Ungridded data or a Cube) and returns
             one new LazyData object with the values as determined by the constraint and kernel objects. The metadata
             for the output LazyData object is copied from the input data object.
        @param points: A list of HyperPoints
        @param data: An UngriddedData object or Cube, or any other object containing metadata that the constraint object can read
        @param constraint: An instance of a Constraint subclass which takes a data object and returns a subset of that data
                            based on it's internal parameters
        @param kernel: An instance of a Kernel subclass which takes a numberof points and returns a single value - This
                            should be full_average - no other kernels currently return multiple values
        @return: One LazyData object for the mean of the constrained values, one for the standard deviation and another
                    for the number of points in the constrained set for which the mean was calculated
        '''
        from jasmin_cis.data_io.ungridded_data import LazyData, UngriddedData, Metadata
        from jasmin_cis.exceptions import ClassNotFoundError

        import numpy as np

        metadata = data.metadata

        if not isinstance(kernel, full_average):
            raise ClassNotFoundError("Invalid kernel specified for this colocator. Should be 'full_average'.")

        # Convert ungridded data to a list of points
        if isinstance(data, UngriddedData):
            data = data.get_non_masked_points()

        logging.info("--> colocating...")

        # Fill will the FillValue from the start
        means = np.zeros(len(points)) + constraint.fill_value
        stddev = np.zeros(len(points)) + constraint.fill_value
        nopoints = np.zeros(len(points)) + constraint.fill_value

        for i, point in enumerate(points):
            con_points = constraint.constrain_points(point, data)
            try:
                means[i], stddev[i], nopoints[i] = kernel.get_value(point, con_points)
            except ValueError:
                pass

        mean_data = LazyData(means, metadata)
        if self.var_name:
            mean_data.metadata._name = self.var_name
        else:
            mean_data.metadata._name = mean_data.name() + '_mean'
        if self.var_long_name: mean_data.metadata.long_name = self.var_long_name
        if self.var_units: mean_data.units = self.var_units
        mean_data.metadata.shape = (len(points),)
        mean_data.metadata.missing_value = constraint.fill_value

        if not self.stddev_name:
            self.stddev_name = mean_data.name()+'_std_dev'
        stddev_data = LazyData(stddev, Metadata(name=self.stddev_name,
                                                long_name='Standard deviation from the mean in '+metadata._name,
                                                shape=(len(points),), missing_value=constraint.fill_value, units=mean_data.units))

        if not self.nopoints_name:
            self.nopoints_name = mean_data.name()+'_no_points'
        nopoints_data = LazyData(nopoints, Metadata(name=self.nopoints_name,
                                                    long_name='Number of points used to calculate the mean of '+metadata._name,
                                                    shape=(len(points),), missing_value=constraint.fill_value, units='1'))

        return [mean_data, stddev_data, nopoints_data]


class DifferenceColocator(Colocator):

    def __init__(self, var_name='', var_long_name='', var_units='', diff_name='difference', diff_long_name=''):
        super(DifferenceColocator, self).__init__()
        self.var_name = var_name
        self.var_long_name = var_long_name
        self.var_units = var_units
        self.diff_name = diff_name
        self.diff_long_name= diff_long_name

    def colocate(self, points, data, constraint, kernel):
        '''
            This colocator takes a list of HyperPoints and a data object (currently either Ungridded data or a Cube) and returns
             one new LazyData object with the values as determined by the constraint and kernel objects. The metadata
             for the output LazyData object is copied from the input data object.
        @param points: A list of HyperPoints
        @param data: An UngriddedData object or Cube, or any other object containing metadata that the constraint object can read
        @param constraint: An instance of a Constraint subclass which takes a data object and returns a subset of that data
                            based on it's internal parameters
        @param kernel: An instance of a Kernel subclass which takes a number of points and returns a single value
        @return: One LazyData object for the colocated data, and another for the difference between that data and the sample data
        '''
        from jasmin_cis.data_io.ungridded_data import LazyData, UngriddedData, Metadata
        import numpy as np

        metadata = data.metadata

        # Convert ungridded data to a list of points
        if isinstance(data, UngriddedData):
            data = data.get_non_masked_points()

        logging.info("--> colocating...")

        # Fill will the FillValue from the start
        values = np.zeros(len(points)) + constraint.fill_value
        difference = np.zeros(len(points)) + constraint.fill_value

        for i, point in enumerate(points):
            con_points = constraint.constrain_points(point, data)
            try:
                values[i] = kernel.get_value(point, con_points)
                difference[i] = values[i] - point.val[0]
            except ValueError:
                pass

        val_data = LazyData(values, metadata)
        if self.var_name: val_data.metadata._name = self.var_name
        if self.var_long_name: val_data.metadata.long_name = self.var_long_name
        if self.var_units: val_data.units = self.var_units
        val_data.metadata.shape = (len(points),)
        val_data.metadata.missing_value = constraint.fill_value

        if not self.diff_long_name: self.diff_long_name = 'Difference between given variable and sampling values'
        diff_data = LazyData(difference, Metadata(name=self.diff_name, long_name=self.diff_long_name, shape=(len(points),),
                                                  missing_value=constraint.fill_value, units=val_data.units))

        return [val_data, diff_data]


class DebugColocator(Colocator):

    def __init__(self, max_vals=1000, print_step=10.0):
        super(DebugColocator, self).__init__()
        self.max_vals = int(max_vals)
        self.print_step = float(print_step)

    def colocate(self, points, data, constraint, kernel):
        # This is the same colocate method as above with extra logging and timing steps. This is useful for debugging
        #  but will be slower than the default colocator.
        from jasmin_cis.data_io.ungridded_data import LazyData, UngriddedData
        import numpy as np
        import math
        from time import time

        metadata = data.metadata

        # Convert ungridded data to a list of points
        if isinstance(data, UngriddedData):
            data = data.get_non_masked_points()

        logging.info("--> colocating...")

        # Only colocate a certain number of points, as a quick test
        short_points = points if len(points)<self.max_vals else points[:self.max_vals-1]

        # We still need to output the full size list, to match the size of the coordinates
        values = np.zeros(len(points)) + constraint.fill_value

        times = np.zeros(len(short_points))
        for i, point in enumerate(short_points):

            t1 = time()

            # colocate using a constraint and a kernel
            con_points = constraint.constrain_points(point, data)
            try:
                values[i] = kernel.get_value(point, con_points)
            except ValueError:
                pass

            # print debug information to screen
            times[i] = time() - t1
            frac, rem = math.modf(i/self.print_step)
            if frac == 0: print str(i) + " - took: " + str(times[i]) + "s" + " -  sample: " + str(point) + " - colocated value: " + str(values[i])

        logging.info("Average time per point: " + str(np.sum(times)/len(short_points)))
        new_data = LazyData(values, metadata)
        new_data.metadata.shape = (len(points),)
        new_data.metadata.missing_value = constraint.fill_value
        return [new_data]


class DummyColocator(Colocator):

    def colocate(self, points, data, constraint, kernel):
        '''
            This colocator does no colocation at all - it just returns the original data values. This might be useful
            if the input data for one variable is already known to be on the same grid as points. This routine could
            check the coordinates are the same but currently does no such check.
        @param points: A list of HyperPoints
        @param data: An UngriddedData object or Cube
        @param constraint: Unused
        @param kernel: Unused
        @return: A single LazyData object
        '''
        from jasmin_cis.data_io.ungridded_data import LazyData

        logging.info("--> colocating...")

        new_data = LazyData(data.data, data.metadata)
        return [new_data]


class DummyConstraint(Constraint):

    def constrain_points(self, point, data):
        # This is a null constraint - all of the points just get passed back
        return data


class SepConstraint(Constraint):

    def __init__(self, h_sep=None, a_sep=None, t_sep=None, fill_value=None):
        from jasmin_cis.exceptions import InvalidCommandLineOptionError

        super(SepConstraint, self).__init__()
        if fill_value is not None:
            try:
                self.fill_value = float(fill_value)
            except ValueError:
                raise InvalidCommandLineOptionError('Seperation Constraint fill_value must be a valid float')
        self.checks = []
        if h_sep is not None:
            try:
                self.h_sep = float(h_sep)
            except ValueError:
                raise InvalidCommandLineOptionError('Seperation Constraint h_sep must be a valid float')
            self.checks.append(self.horizontal_constraint)
        if a_sep is not None:
            try:
                self.a_sep = float(a_sep)
            except:
                raise InvalidCommandLineOptionError('Seperation Constraint a_sep must be a valid float')
            self.checks.append(self.alt_constraint)
        if t_sep is not None:
            from jasmin_cis.time_util import parse_datetimestr_delta_to_float_days
            try:
                self.t_sep = parse_datetimestr_delta_to_float_days(t_sep)
            except ValueError as e:
                raise InvalidCommandLineOptionError(e)
            self.checks.append(self.time_constraint)

    def time_constraint(self, point, ref_point):
        return point.time_sep(ref_point) < self.t_sep

    def alt_constraint(self, point, ref_point):
        return point.alt_sep(ref_point) < self.a_sep

    def horizontal_constraint(self, point, ref_point):
        return point.haversine_dist(ref_point) < self.h_sep

    def constrain_points(self, ref_point, data):
        from jasmin_cis.data_io.hyperpoint import HyperPointList
        con_points = HyperPointList()
        for point in data:
            if all(check(point, ref_point) for check in self.checks):
                con_points.append(point)
        return con_points


class mean(Kernel):

    def get_value(self, point, data):
        '''
            Colocation using the mean of any points left after a constraint.
        '''
        from numpy import mean
        values = data.vals
        if len(values) == 0: raise ValueError
        return mean(values)


class full_average(Kernel):

    def get_value(self, point, data):
        '''
            Colocation using the mean of any points left after a constraint. Also returns the standard
             deviation and the number of points
        '''
        from numpy import mean, std
        values = data.vals
        num_values = len(values)
        if num_values == 0: raise ValueError
        return (mean(values), std(values), num_values)


class nn_horizontal(Kernel):

    def get_value(self, point, data):
        '''
            Colocation using nearest neighbours along the face of the earth where both points and
              data are a list of HyperPoints. The default point is the first point.
        '''
        if len(data) == 0: raise ValueError
        nearest_point = data[0]
        for data_point in data[1:]:
            if point.compdist(nearest_point, data_point): nearest_point = data_point
        return nearest_point.val[0]


class nn_vertical(Kernel):

    def get_value(self, point, data):
        '''
            Colocation using nearest neighbours in altitude, where both points and
              data are a list of HyperPoints. The default point is the first point.
        '''
        if len(data) == 0: raise ValueError
        nearest_point = data[0]
        for data_point in data[1:]:
            if point.compalt(nearest_point, data_point): nearest_point = data_point
        return nearest_point.val[0]


class nn_time(Kernel):

    def get_value(self, point, data):
        '''
            Colocation using nearest neighbours in time, where both points and
              data are a list of HyperPoints. The default point is the first point.
        '''
        if len(data) == 0: raise ValueError
        nearest_point = data[0]
        for data_point in data[1:]:
            if point.comptime(nearest_point, data_point): nearest_point = data_point
        return nearest_point.val[0]


class nn_gridded(Kernel):
    def get_value(self, point, data):
        '''
            Co-location routine using nearest neighbour algorithm optimized for gridded data.
             This calls out to iris to do the work.
        '''
        from iris.analysis.interpolate import nearest_neighbour_data_value
        return nearest_neighbour_data_value(data, point.coord_tuple)


class li(Kernel):
    def get_value(self, point, data):
        '''
            Co-location routine using iris' linear interpolation algorithm. This only makes sense for gridded data.
        '''
        from iris.analysis.interpolate import linear
        return linear(data, point.get_coord_tuple()).data



'''
Module for writing data to NetCDF files
'''
from netCDF4 import Dataset
import logging

types = {'int16': "i2",
         'int32': "i4",
         'int64': "i8",
         'float32': "f4",
         'float64': "f8"}

index_name = 'pixel_number'


def __add_metadata(var, data):
    if data.standard_name: var.standard_name = data.standard_name
    if data.units: var.units = str(data.units)
    if data.long_name: var.long_name = data.long_name
    if data.metadata.range : var.valid_range = data.metadata.range
    if data.metadata.missing_value : var.missing_value = data.metadata.missing_value
    if data.metadata.calendar : var.calendar = data.metadata.calendar
    if data.metadata.history : var.history = data.metadata.history
    var.shape = data.shape
    return var


def __get_missing_value(coord):
    f = coord.metadata.missing_value
    if not f and f !=0:
        f = None
    return f


def __create_variable(nc_file, data):
    from jasmin_cis.exceptions import InconsistentDimensionsError
    logging.info("Creating variable: " + data.standard_name + "("+index_name+")" + " " + types[str(data.data.dtype)])
    var = nc_file.createVariable(data.standard_name, types[str(data.data.dtype)], index_name, fill_value=__get_missing_value(data))
    var = __add_metadata(var, data)
    try:
        var[:] = data.data.flatten()
    except IndexError as e:
        raise InconsistentDimensionsError(str(e)+"\nInconsistent dimensions in output file, unable to write "
                                                 ""+data.standard_name+" to file (it's shape is "+str(data.shape)+").")

    return var


def __create_index(nc_file, length):
    import numpy as np
    dimension = nc_file.createDimension(index_name, length)
    dimensions = ( index_name, )
    var = nc_file.createVariable(index_name, np.int32, dimensions)

    var.valid_range = (0, length)
    var[:] = np.arange(length)

    return dimensions


def write(data_object, filename):
    """

    @param data_object:
    @param filename:
    @return:
    """
    write_coordinates(data_object.coords(), filename)
    add_data_to_file(data_object, filename)


def write_coordinates(coords, filename):
    """

    @param coords:
    @param filename:
    @return:
    """
    netcdf_file = Dataset(filename, 'w', format="NETCDF4_CLASSIC")
    index_dim = __create_index(netcdf_file, len(coords[0].data.flatten()))
    for data in coords:
        coord = __create_variable(netcdf_file, data)
    netcdf_file.close()


def add_data_to_file(data_object, filename):
    """

    @param data_object:
    @param filename:
    @return:
    """
    netcdf_file = Dataset(filename, 'a', format="NETCDF4_CLASSIC")
    var = __create_variable(netcdf_file, data_object)
    netcdf_file.close()



"""Tests for parse_datetime module
"""
import datetime
from nose.tools import istest, raises
from jasmin_cis.parse_datetime import (parse_datetime, parse_as_number_or_datetime, find_last_day_of_month,
                                       convert_datetime_components_to_datetime)


class MockParserError(Exception):
    pass


class MockParser(object):
    def error(self, message):
        raise MockParserError(message)


# Tests for parse_datetime
@istest
def parse_datetime_can_parse_year():
    parser = MockParser()
    dt = parse_datetime('2010', 'date/time arg', parser)
    assert(dt == [2010])


@istest
def parse_datetime_can_parse_year_month():
    parser = MockParser()
    dt = parse_datetime('2010-07', 'date/time arg', parser)
    print dt
    assert(dt == [2010, 7])


@istest
def parse_datetime_can_parse_date():
    parser = MockParser()
    dt = parse_datetime('2010-07-01', 'date/time arg', parser)
    assert(dt == [2010, 7, 1])


@istest
def parse_datetime_can_parse_date_hour():
    parser = MockParser()
    dt = parse_datetime('2010-07-01T13', 'date/time arg', parser)
    assert(dt == [2010, 7, 1, 13])


@istest
def parse_datetime_can_parse_date_hour_min():
    parser = MockParser()
    dt = parse_datetime('2010-07-01T13:27', 'date/time arg', parser)
    assert(dt == [2010, 7, 1, 13, 27])


@istest
def parse_datetime_can_parse_date_hour_min_sec():
    parser = MockParser()
    dt = parse_datetime('2010-07-01T13:27:43', 'date/time arg', parser)
    assert(dt == [2010, 7, 1, 13, 27, 43])


@istest
def parse_datetime_can_parse_date_hour_min_sec_no_leading_zeros():
    parser = MockParser()
    dt = parse_datetime('2010-3-4T5:6:7', 'date/time arg', parser)
    assert(dt == [2010, 3, 4, 5, 6, 7])


@istest
def parse_datetime_can_parse_date_time_with_space_separator():
    parser = MockParser()
    dt = parse_datetime('2010-07-01 13:27:43', 'date/time arg', parser)
    assert(dt == [2010, 7, 1, 13, 27, 43])


@istest
def parse_datetime_can_parse_date_time_with_colon_separator():
    parser = MockParser()
    dt = parse_datetime('2010-07-01:13:27:43', 'date/time arg', parser)
    assert(dt == [2010, 7, 1, 13, 27, 43])


@istest
def parse_datetime_passed_error_message_to_parser():
    parser = MockParser()
    name = 'date/time arg'
    try:
        dt = parse_datetime('2X10', name, parser)
    except MockParserError as e:
        assert e.message.index(name) > 0


# parse_datetime: Parse errors
@istest
@raises(MockParserError)
def parse_datetime_raises_error_if_invalid_character_in_year():
    parser = MockParser()
    dt = parse_datetime('2X10', 'date/time arg', parser)


@istest
@raises(MockParserError)
def parse_datetime_raises_error_if_time_but_incomplete_date():
    parser = MockParser()
    dt = parse_datetime('2010-10T12:00', 'date/time arg', parser)


@istest
@raises(MockParserError)
def parse_datetime_raises_error_if_too_many_date_components():
    parser = MockParser()
    dt = parse_datetime('2010-10-05-06', 'date/time arg', parser)


@istest
@raises(MockParserError)
def parse_datetime_raises_error_if_too_many_time_components():
    parser = MockParser()
    dt = parse_datetime('2010-10-05T12:01:02:03', 'date/time arg', parser)


# parse_datetime: Strings that parse correctly but correspond to invalid date/times
@istest
@raises(MockParserError)
def parse_datetime_raises_error_if_invalid_month():
    parser = MockParser()
    dt = parse_datetime('2010-13', 'date/time arg', parser)


@istest
@raises(MockParserError)
def parse_datetime_raises_error_if_invalid_day():
    parser = MockParser()
    dt = parse_datetime('2010-06-31', 'date/time arg', parser)


# Tests for parse_as_number_or_datetime
@istest
def parse_as_number_or_datetime_can_parse_date_as_datetime():
    parser = MockParser()
    dt = parse_as_number_or_datetime('2010-07-01', 'date/time arg', parser)
    assert(dt == [2010, 7, 1])


@istest
def parse_as_number_or_datetime_can_parse_integer():
    parser = MockParser()
    dt = parse_as_number_or_datetime('2010', 'limit arg', parser)
    assert(dt == 2010)


@istest
def parse_as_number_or_datetime_can_parse_float():
    parser = MockParser()
    dt = parse_as_number_or_datetime('12.345', 'limit arg', parser)
    assert(dt == 12.345)


# Tests for find_last_day_of_month
@istest
def find_last_day_of_month_finds_day_for_dec_2010():
    day = find_last_day_of_month(2010, 12)
    assert(day == 31)


@istest
def find_last_day_of_month_finds_day_for_feb_2000():
    day = find_last_day_of_month(2000, 2)
    assert(day == 29)


# Tests for convert_datetime_components_to_datetime
@istest
def convert_datetime_components_to_datetime_can_convert_year_as_lower_limit():
    dt = convert_datetime_components_to_datetime([2000], True)
    assert(dt == datetime.datetime(2000, 1, 1, 0, 0, 0))


@istest
def convert_datetime_components_to_datetime_can_convert_year_int_as_lower_limit():
    dt = convert_datetime_components_to_datetime(2000, True)
    assert(dt == datetime.datetime(2000, 1, 1, 0, 0, 0))


@istest
def convert_datetime_components_to_datetime_can_convert_year_as_upper_limit():
    dt = convert_datetime_components_to_datetime([2000], False)
    assert(dt == datetime.datetime(2000, 12, 31, 23, 59, 59))


@istest
def convert_datetime_components_to_datetime_can_convert_year_int_as_upper_limit():
    dt = convert_datetime_components_to_datetime(2000, False)
    assert(dt == datetime.datetime(2000, 12, 31, 23, 59, 59))


@istest
def convert_datetime_components_to_datetime_can_convert_year_month_as_lower_limit():
    dt = convert_datetime_components_to_datetime([1990, 6], True)
    assert(dt == datetime.datetime(1990, 6, 1, 0, 0, 0))


@istest
def convert_datetime_components_to_datetime_can_convert_year_month_as_upper_limit():
    dt = convert_datetime_components_to_datetime([1990, 6], False)
    assert(dt == datetime.datetime(1990, 6, 30, 23, 59, 59))


@istest
def convert_datetime_components_to_datetime_can_convert_year_month_day_as_lower_limit():
    dt = convert_datetime_components_to_datetime([1990, 6, 7], True)
    assert(dt == datetime.datetime(1990, 6, 7, 0, 0, 0))


@istest
def convert_datetime_components_to_datetime_can_convert_year_month_day_as_upper_limit():
    dt = convert_datetime_components_to_datetime([1990, 6, 7], False)
    assert(dt == datetime.datetime(1990, 6, 7, 23, 59, 59))


@istest
def convert_datetime_components_to_datetime_can_convert_date_hour_as_lower_limit():
    dt = convert_datetime_components_to_datetime([1990, 6, 7, 18], True)
    assert(dt == datetime.datetime(1990, 6, 7, 18, 0, 0))


@istest
def convert_datetime_components_to_datetime_can_convert_date_hour_as_upper_limit():
    dt = convert_datetime_components_to_datetime([1990, 6, 7, 18], False)
    assert(dt == datetime.datetime(1990, 6, 7, 18, 59, 59))


@istest
def convert_datetime_components_to_datetime_can_convert_date_hour_min_as_lower_limit():
    dt = convert_datetime_components_to_datetime([1990, 6, 7, 6, 30], True)
    assert(dt == datetime.datetime(1990, 6, 7, 6, 30, 0))


@istest
def convert_datetime_components_to_datetime_can_convert_date_hour_min_as_upper_limit():
    dt = convert_datetime_components_to_datetime([1990, 6, 7, 6, 30], False)
    assert(dt == datetime.datetime(1990, 6, 7, 6, 30, 59))


@istest
def convert_datetime_components_to_datetime_can_convert_date_hour_min_sec_as_lower_limit():
    dt = convert_datetime_components_to_datetime([1990, 6, 7, 12, 15, 45], True)
    assert(dt == datetime.datetime(1990, 6, 7, 12, 15, 45))


@istest
def convert_datetime_components_to_datetime_can_convert_date_hour_min_sec_as_upper_limit():
    dt = convert_datetime_components_to_datetime([1990, 6, 7, 12, 15, 45], False)
    assert(dt == datetime.datetime(1990, 6, 7, 12, 15, 45))


@istest
@raises(ValueError)
def convert_datetime_components_to_datetime_raises_error_if_invalid_date():
    dt = convert_datetime_components_to_datetime([2000, 6, 31], True)


@istest
@raises(ValueError)
def convert_datetime_components_to_datetime_raises_error_if_invalid_time():
    dt = convert_datetime_components_to_datetime([2000, 6, 30, 12, 30, 60], True)


if __name__ == '__main__':
    import nose
    nose.runmodule()



"""
Module containing NetCDF file reading functions
"""


def get_netcdf_file_variables(filename):
    '''
    Get all the variables from a NetCDF file
    
    @param filename: The filename of the file to get the variables from
    @return: An OrderedDict containing the variables from the file
    '''
    from netCDF4 import Dataset    
    f = Dataset(filename)
    return f.variables

def read_many_files(filenames, usr_variables, dim=None):
    """
    Reads a single Variable from many NetCDF files.

        @param filenames: A list of NetCDF filenames to read, or a string with wildcards.
        @param usr_variables: A list of variable (dataset) names to read from the
                       files. The names must appear exactly as in in the NetCDF file.
        @param dim: The name of the dimension on which to aggregate the data. None is the default
                     which tries to aggregate over the unlimited dimension
        @return: A list of variable instances constructed from all of the input files
    """
    from netCDF4 import MFDataset
    from jasmin_cis.exceptions import InvalidVariableError

    if not isinstance(usr_variables,list):
        usr_variables = [usr_variables]

    try:
        datafile = MFDataset(filenames, aggdim=dim)
    except RuntimeError as e:
        raise IOError(e)

    data = {}
    for variable in usr_variables:
        # Get data.
        try:
            data[variable] = datafile.variables[variable]
        except:
            raise InvalidVariableError

    return data


def read_many_files_individually(filenames, usr_variables):
    """
    Read multiple Variables from many NetCDF files manually - i.e. not with MFDataset as this doesn't alays work

        @param filenames: A list of NetCDF filenames to read, or a string with wildcards.
        @param usr_variables: A list of variable (dataset) names to read from the
                       files. The names must appear exactly as in in the NetCDF file.
        @return: A dictionary of lists of variable instances constructed from all of the input files with the variable
                    name as the key
    """
    from jasmin_cis.utils import add_element_to_list_in_dict

    if not isinstance(usr_variables,list):
        usr_variables = [usr_variables]

    var_data = {}

    for filename in filenames:

        var_dict = read(filename, usr_variables)
        for var in var_dict.keys():
            add_element_to_list_in_dict(var_data, var, var_dict[var])

    return var_data


def read(filename, usr_variables):
    """
    Reads a Variable from a NetCDF file

        @param filename: The name (with path) of the NetCDF file to read.
        @param usr_variables: A variable (dataset) name to read from the
                       files. The name must appear exactly as in in the NetCDF file.
        @return: A Variable instance constructed from  the input file

    """
    from netCDF4 import Dataset
    from jasmin_cis.exceptions import InvalidVariableError

    if not isinstance(usr_variables,list):
        usr_variables = [usr_variables]

    try:
        datafile = Dataset(filename)
    except RuntimeError as e:
        raise IOError(str(e))

    data = {}
    for variable in usr_variables:
        # Get data.
        try:
            data[variable] = datafile.variables[variable]
        except:
            raise InvalidVariableError(variable+' could not be found in '+filename)

    return data

def get_metadata(var):
    '''
    Retrieves all metadata

    @param var: the Variable to read metadata from
    @return: A metadata object
    '''
    from jasmin_cis.data_io.ungridded_data import Metadata

    standard_name = getattr(var,'standard_name',"")
    missing_value = find_missing_value(var)
    long_name = getattr(var,'long_name',"")
    units = getattr(var, 'units', "")

    history = getattr(var, "history", "")
    shape = getattr(var, "shape", None)
    if shape is None:
        try:
            shape=(var._recLen[0],)
        except AttributeError:
            shape = ()

    metadata = Metadata(var._name, standard_name, long_name, units=units, missing_value=missing_value, shape=shape, history=history)

    return metadata

def find_missing_value(var):
    try:
        missing_value = var.missing_value
    except AttributeError:
        try:
            missing_value = var._FillValue
        except AttributeError:
            missing_value = None
    return missing_value

def get_data(var, calipso_scaling=False):
    """
    Reads raw data from a NetCDF.Variable instance.

        @param var: The specific Variable instance to read
        @return:  A numpy maskedarray. Missing values are False in the mask.
    """
    from jasmin_cis.utils import create_masked_array_for_missing_data
    # Note that this will automatically apply any specified scalings and
    #  return a masked array based on _FillValue
    data = var[:]

    return data



from jasmin_cis.data_io import hdf_sd as hdf_sd, hdf_vd
import jasmin_cis.utils as utils
import logging

def get_hdf4_file_variables(filename, data_type):
    """
    Get all variables from a file containing ungridded data.
    Concatenate variable from both VD and SD data

    @param filename: The filename of the file to get the variables from
    @param data_type: String representing the HDF data type, i.e. 'VD' or 'SD'. if None, both are computed.
    """

    SD_vars = VD_vars = None

    if data_type is None:
        SD_vars = hdf_sd.get_hdf_SD_file_variables(filename)
        VD_vars = hdf_vd.get_hdf_VD_file_variables(filename)
    else:
        if data_type.lower() == 'SD'.lower() or data_type.lower() == "all".lower():
            SD_vars = hdf_sd.get_hdf_SD_file_variables(filename)

        if data_type.lower() == 'VD'.lower() or data_type.lower() == "all".lower():
            VD_vars = hdf_vd.get_hdf_VD_file_variables(filename)

    return SD_vars, VD_vars


def get_hdf4_file_metadata(filename):
    """
    This returns a dictionary of file attributes, which often contains metadata information
    about the whole file. The value of each attribute can simply be a big string
    which will often need to be parsed manually thereafter.
    @param filename
    @return: dictionary of string attributes
    """
    from pyhdf.SD import SD
    return SD(filename).attributes()

def __read_hdf4(filename,variables):
    '''
        A wrapper method for reading raw data from hdf4 files. This returns a dictionary of io handles
         for each VD and SD data types.

        @param filename:     A name of a file to read
        @param variables:    List of variables to read from the files

        @return (sds_dict, vds_dict) A tuple of dictionaries, one for sds objects and another for vds
    '''
    from jasmin_cis.exceptions import InvalidVariableError
    from pyhdf.error import HDF4Error

    if not isinstance(variables,list): variables = [ variables ]

    try:
        sds_dict = hdf_sd.read(filename,variables)

        # remove the variables identified as SD (i.e. the keys in sds_dict)
        # no need to try looking for them as VD variable
        # AND this can cause a crash in some version/implementations of the core HDF4 libraries!

        # First create a copy of the list in order for the original list to be left intact when elements are removed from it
        # This enables the original list to be used when many files are read
        vdvariables = list(variables)
        for sds_dict_key in sds_dict:
            vdvariables.remove(sds_dict_key)

        vds_dict = hdf_vd.read(filename, vdvariables)
    except HDF4Error as e:
        joined_up_message = "".join(e)
        raise IOError(joined_up_message)

    for variable in variables:
        if variable not in sds_dict and variable not in vds_dict:
            raise InvalidVariableError("Could not find " + variable + " in file: " + filename)

    return sds_dict, vds_dict


def read(filenames,variables):

    sdata = {}
    vdata = {}

    for filename in filenames:

        logging.debug("reading file: " + filename)

        # reading in all variables into a 2 dictionaries:
        # sdata, key: variable name, value: list of sds
        # vdata, key: variable name, value: list of vds
        sds_dict, vds_dict = __read_hdf4(filename,variables)
        for var in sds_dict.keys():
            utils.add_element_to_list_in_dict(sdata,var,sds_dict[var])
        for var in vds_dict.keys():
            utils.add_element_to_list_in_dict(vdata,var,vds_dict[var])

    return sdata,vdata

def read_data(data_dict, data_type, missing_values=None):
    if data_type=='VD':
        out =  utils.concatenate([hdf_vd.get_data(i, missing_values=missing_values) for i in data_dict ])
    if data_type=='SD':
        out =  utils.concatenate([hdf_sd.get_data(i, missing_values=missing_values) for i in data_dict ])
    return out

def read_metadata(data_dict, data_type):
    if data_type=='VD':
        out = hdf_vd.get_metadata(data_dict[0])
    if data_type=='SD':
        out = hdf_sd.get_metadata(data_dict[0])
    return out







import logging

def get_aeronet_file_variables(filename):
    import linecache
    defaultdeletechars = """~!@#$%^&*()-=+~\|]}[{';: /?.>,<"""
    vars = linecache.getline(filename, 5).split(",")
    for i in range(0, len(vars)):
        for char in defaultdeletechars:
            vars[i] = vars[i].replace(char, "")
    vars_dict = {}
    for var in vars:
        vars_dict[var] = var
    return vars_dict


def load_multiple_aeronet(fnames, variables=None):
    from jasmin_cis.utils import add_element_to_list_in_dict, concatenate

    adata = {}

    for filename in fnames:
        logging.debug("reading file: " + filename)

        # reading in all variables into a dictionary:
        # a_dict, key: variable name, value: list of masked arrays
        a_dict = load_aeronet(filename,variables)
        for var in a_dict.keys():
            add_element_to_list_in_dict(adata,var,a_dict[var])

    for var in adata.keys():
        adata[var] = concatenate(adata[var])

    return adata


def load_aeronet(fname, variables=None):
    """
    loads aeronet lev 2.0 csv file.

        Originally from http://code.google.com/p/metamet/
        License: GNU GPL v3

    @param fname: data file name
    @param keep_fields: A list of variables to return
    @return: A
    """
    import numpy as np
    from numpy import ma
    from datetime import datetime, timedelta

    std_day = datetime(1900,1,1,0,0,0)
    def date2daynum(datestr):
        the_day = datetime.strptime(datestr, '%d:%m:%Y')
        return float((the_day - std_day).days)

    def time2seconds(timestr):
        h, m, s = [int(t) for t in timestr.split(':')]
        return float(h * 3600 + m * 60 + s)

    def daynum_seconds2datetime(daynum, seconds):
        return std_day + timedelta(days=int(daynum), seconds=int(seconds))

    try:
        rawd = np.genfromtxt(fname, skip_header=4, delimiter=',', names=True,
                             converters={0:date2daynum, 1:time2seconds}, missing_values='N/A', usemask=True)
    except (StopIteration, IndexError) as e:
        raise IOError(e)

    lend = len(rawd)
    dates = np.zeros(lend, dtype='O')
    for i in xrange(lend):
        dates[i] = daynum_seconds2datetime(rawd['Dateddmmyy'][i], rawd['Timehhmmss'][i])

    metadata = get_file_metadata(fname)
    lon = np.zeros(lend) + float(metadata.misc[2][1].split("=")[1])
    lat = np.zeros(lend) + float(metadata.misc[2][2].split("=")[1])
    alt = np.zeros(lend) + float(metadata.misc[2][3].split("=")[1])

    data_dict = {}
    if variables is not None:
        for key in variables:
            data_dict[key] = rawd[key]

    data_dict["datetime"] = ma.array(dates)
    data_dict["longitude"] = ma.array(lon)
    data_dict["latitude"] = ma.array(lat)
    data_dict["altitude"] = ma.array(alt)

    return data_dict


def get_file_metadata(filename, variable = '', shape = None):
    file = open(filename)
    from jasmin_cis.data_io.ungridded_data import Metadata
    if variable is None: variable = ''
    metadata = Metadata(name = variable, long_name = variable, shape = shape)
    lines = []
    for i in range(0, 4):
        lines.append(file.readline().replace("\n","").split(","))
    metadata.misc = lines
    return metadata


'''
module to test the reading of ungridded data files
'''
from nose.tools import istest, eq_, raises, nottest

@nottest
def from_points_array_returns_valid_ungridded_data_object():
    pass

@nottest
def can_create_ungridded_data_object_from_numpy_array():
    pass

@nottest
def can_create_ungridded_data_object_from_hdf_vd_obj():
    pass

@nottest
def can_create_ungridded_data_object_from_hdf_sd_obj():
    pass

@nottest
def can_create_ungridded_data_object_from_netcdf_data_obj():
    pass


from setuptools import setup, find_packages, Command
from pkg_resources import require, DistributionNotFound, VersionConflict
from distutils.spawn import spawn, find_executable
import sys, os, os.path

# Extension classes and functions to add custom command
#======================================================


class check_dep(Command):
    """
    Command to check that the required dependencies are installed on the system
    """
    description = "Checks that the required dependencies are installed on the system"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):

        deps = ["matplotlib>=1.2.0","pyke","cartopy","Shapely","netcdf4",
                "numpy","scipy","nose","iris"]
        for dep in deps:
            try:
                require(dep)
                print dep + " ...[ok]"
            except (DistributionNotFound, VersionConflict):
                print dep + "... MISSING!" 

class gen_doc(Command):
    """
    Command to generate the API reference using epydoc
    """
    description = "Generates the API reference using epydoc"
    user_options = []

    def initialize_options(self):
        require("epydoc")

    def finalize_options(self):
        pass

    def run(self):
        #raise NotImplementedError, "not implemented yet"  

        if not find_executable('epydoc'):
            print "Could not find epydoc on system"
            return  

        # create output directory if does not exists
        output_dir = "doc/api"
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        from subprocess import call
        call(["epydoc", "--html", "--no-sourcecode", "-o", output_dir, "*"])

        import webbrowser
        webbrowser.open(os.path.join(output_dir,"index.html"))


# Metadata description
#=====================

# Extract long-description from README
here = os.path.dirname(__file__)
README = open(os.path.join(here, 'README')).read()

setup(
    name='jasmin_cis',
    version='0.6.1',
    description='JASMIN Community Inter-comparison Suite',
    long_description=README,
    maintainer='Stephen Pascoe',
    maintainer_email='Stephen.Pascoe@stfc.ac.uk',
    url='http://proj.badc.rl.ac.uk/cedaservices/wiki/JASMIN/CommunityIntercomparisonSuite',
    classifiers = [
        'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
        'Topic :: Scientific/Engineering :: Atmospheric Science',
        'Topic :: Scientific/Engineering :: Visualization',
        'Environment :: Console',
        ],

    packages=find_packages(),
    package_data={'':['logging.conf']},

    scripts = ['bin/cis'],
    
    cmdclass={"gendoc": gen_doc, "checkdep":check_dep}
)





from jasmin_cis.data_io.products.AProduct import get_data

filenames = ["/home/daniel/Aeronet_Files/Data/920801_091128_Abracos_Hill_small.lev20"]
data_object1 = get_data(filenames, 'AOT_1020',"Aeronet")
data_object2 = get_data(filenames, 'AOT_870',"Aeronet")
data_object3 = get_data(filenames, 'AOT_675',"Aeronet")
data_object4 = get_data(filenames, 'AOT_500',"Aeronet")
data_object5 = get_data(filenames, 'Watercm',"Aeronet")

from plotting.plot import Plotter

Plotter([data_object1, data_object2, data_object3, data_object4, data_object5])


"""
Module containing hdf file utility functions for the VD object
"""
import numpy as np
from pyhdf.HDF import *
from pyhdf.VS import *
from collections import namedtuple
import logging
from jasmin_cis.utils import create_masked_array_for_missing_values


class VDS(namedtuple('VDS',['filename','variable'])):
    pass

def get_hdf_VD_file_variables(filename):
    '''
    Get all the variables from an HDF VD file

    args:
        filename: The filename of the file to get the variables from

    returns:
        An OrderedDict containing the variables from the file
    '''
    variables = None
    try:
        # Open file
        datafile = HDF(filename)
        vs =  datafile.vstart()
        # List of required variable names
        names = vs.vdatainfo()
        # This returns a list of tuples, so convert into a dictionary for easy lookup
        variables = {}
        for var in names:
            variables[var[0]] = var[1:]
        # Close file
        vs.end()
        datafile.close()
    except:
        logging.error("Error while reading VD data")

    return variables

def read(filename, variables=None, datadict=None):

    if datadict == None:
        datadict = {}

    if not isinstance(variables,list): variables = [ variables ]

    datafile = HDF(filename)
    vs = datafile.vstart()

    for variable in variables:
        try:
            vd = vs.attach(variable)
            vd.detach()
            datadict[variable] = VDS(filename, variable)
        except:
            # ignore variable that failed
            pass

    vs.end()
    datafile.close()

    return datadict

def get_data(vds, first_record=False, missing_values=None):

    # get file and variable reference from tuple
    filename = vds.filename
    variable = vds.variable

    try:
        datafile = HDF(filename)
    except HDF4Error as e:
        raise IOError(e)

    vs = datafile.vstart()

    if first_record:
        vd = vs.attach(vs.next(-1))
        vd.setfields(variable)
        data = vd.read()
    else:
        # get data for that variable
        vd = vs.attach(variable)
        data = vd.read(nRec = vd.inquire()[0])

    # create numpy array from data
    data = np.array(data).flatten()

    # dealing with missing data
    if missing_values is None:
        missing_values = [__get_attribute_value(vd, 'missing')]

    data = create_masked_array_for_missing_values(data,missing_values)

    #np.set_printoptions(threshold=np.nan)
    #print data

    # detach and close
    vd.detach()
    vs.end()
    datafile.close()

    return data

def get_metadata(vds):
    from jasmin_cis.data_io.ungridded_data import Metadata

    # get file and variable reference from tuple
    filename = vds.filename
    variable = vds.variable

    datafile = HDF(filename)
    vs =  datafile.vstart()

    # get data for that variable
    vd = vs.attach(variable)

    name = variable
    long_name = __get_attribute_value(vd, 'long_name')
    shape = [len(vd.read(nRec = vd.inquire()[0]))] #VD data are always 1D, so the shape is simply the length of the data vector
    units = __get_attribute_value(vd, 'units')
    valid_range = __get_attribute_value(vd, 'valid_range')
    factor = __get_attribute_value(vd, 'factor')
    offset = __get_attribute_value(vd, 'offset')
    missing = __get_attribute_value(vd, 'missing')

    # put the whole dictionary of attributes into 'misc'
    # so that other metadata of interest can still be retrieved if need be
    misc = vd.attrinfo()

    metadata = Metadata( name=name, long_name=long_name, shape=shape, units=units, range=valid_range,
                            factor=factor, offset=offset, missing_value=missing, misc=misc)

    # detach and close
    vd.detach()
    vs.end()
    datafile.close()

    return metadata

def __get_attribute_value(vd, name):

    val = vd.attrinfo().get(name,None)
    # if the attribute is not present
    if val is None:
        return val
    else:
        #attrinfo() returns a tuple in which the value of interest is the 3rd item, hence the '[2]'
        return val[2]


from plotting.plot import Plotter
from jasmin_cis.data_io.Coord import CoordList, Coord
from jasmin_cis.data_io.ungridded_data import UngriddedData, Metadata
from numpy import array

x = Coord(array([1,2,3,4,5,6,7,8,9]), Metadata('fgfgf'),'x')
y = Coord(array([1,2,3,4,5,6,7,8,9]), Metadata('fgfghghg'),'y')
coords = CoordList([x, y])
data = array([1,2,3,4,5,6,7,8,109])
dataobject = UngriddedData(data, Metadata(standard_name='rain', long_name="TOTAL RAINFALL RATE: LS+CONV KG/M2/S", units="kg m-2 s-1", missing_value=-999), coords)

Plotter([dataobject], plot_type="scatter", logv = False, itemwidth = 400, datagroups = [{"itemstyle" : None, "color" : None, "label" : "Line 1"}])




import collections

def add_element_to_list_in_dict(my_dict,key,value):
    try:
        my_dict[key].append(value)
    except KeyError:
        my_dict[key] = [value]


def concatenate(arrays, axis=0):
    from numpy.ma import MaskedArray
    from numpy import ndarray
    if isinstance(arrays[0], MaskedArray):
        from numpy.ma import concatenate
    elif isinstance(arrays[0], ndarray):
        from numpy import concatenate

    res = arrays[0]

    if len(arrays) > 1:
        for array in arrays[1:]:
            res = concatenate((res,array),axis)

    return res

def calculate_histogram_bin_edges(data, axis, user_range, step, log_scale = False):
    '''
    @param data: A numpy array
    @param axis: The axis on which the data will be plotted. Set to "x" for histogram2d
    @param user_range: A dictionary containing the min and max values for the edges specified by the user. The data min and max is used if the user did not specify
    @param step: The distance between each bin edge/the width of each bin
    @return: An array containing a list of bin edges (i.e. when each bin starts and ends)
    '''
    from decimal import Decimal
    from numpy import array, append, logspace, log10
    import logging
    import sys

    min_val = user_range.get(axis + "min", data.min())
    max_val = user_range.get(axis + "max", data.max())

    val_range = float(max_val - min_val)

    if step is None:
        step = val_range / 10.0
        user_specified_step = False
    else:
        user_specified_step = True

    x = float(Decimal(str(val_range)) % Decimal(str(step)))

    stop = max_val + step if  x < 1.e-7 else max_val # 1.e-7 is very close to 0

    bin_edges = array([])
    i = min_val
    if not log_scale or user_specified_step:
        while abs(i - stop) >= sys.float_info.min and i < stop:
            if not user_specified_step and len(bin_edges) == 11:
                break
            bin_edges = append(bin_edges, i)

            i += step
    else:
        bin_edges = logspace(log10(min_val), log10(stop), num=11)

    logging.debug(axis + " axis bin edges: " + str(bin_edges))
    return bin_edges

def expand_1d_to_2d_array(array_1d,length,axis=0):
    '''
    General utility routine to 'extend a 1D array into a 2D array
    by duplicating the data along a given 'axis' (default is 0)
    of size 'length'.

    Examples
    --------
    >>> a = np.array([1,2,3,4])
    >>> expand_1d_to_2d_array(a, 4, axis=0)
    [[1 2 3 4]
     [1 2 3 4]
     [1 2 3 4]
     [1 2 3 4]]

    >>> a = np.array([1,2,3,4])
    >>> expand_1d_to_2d_array(a, 4, axis=1)
    [[1 1 1 1]
     [2 2 2 2]
     [3 3 3 3]
     [4 4 4 4]]

    @param array_1d:
    @param length:
    @param axis:
    @return:
    '''
    import numpy as np

    if axis==0:
        array_2d = np.repeat(array_1d[np.newaxis,:],length,0)
    else:
        array_2d = np.repeat(array_1d[:,np.newaxis],length,1)

    return array_2d

def create_masked_array_for_missing_data(data, missing_val):
    import numpy.ma as ma
    return ma.array(data, mask=data==missing_val, fill_value=missing_val)

def create_masked_array_for_missing_values(data, missing_values):
    import numpy.ma as ma

    mdata = data
    for missing_value in missing_values:
        mdata = ma.masked_where(missing_value==mdata,mdata)

    return mdata

def unpack_data_object(data_object, x_variable, y_variable):
    '''
    @param data_object    A cube or an UngriddedData object
    @return A dictionary containing x, y and data as numpy arrays
    '''
    from iris.cube import Cube
    import iris.plot as iplt
    import iris
    import logging
    import numpy as np
    from mpl_toolkits.basemap import addcyclic

    def __get_coord(data_object, variable, data):
        from iris.exceptions import CoordinateNotFoundError

        if variable == data_object.name() or variable == "default":
            return data
        else:
            if variable.startswith("search:"):
                number_of_points = float(variable.split(":")[1])
                for coord in data_object.coords():
                    if coord.shape[0] == number_of_points:
                        break
            else:
                try:
                    coord = data_object.coord(name=variable)
                except CoordinateNotFoundError:
                    return None
            if isinstance(data_object, Cube):
                return coord.points
            else:
                return coord.data

    no_of_dims = len(data_object.shape)

    data = data_object.data #ndarray

    x = __get_coord(data_object, x_variable, data)
    y = __get_coord(data_object, y_variable, data)

    if np.array_equal(y, data) or np.array_equal(y, x):
        y = None

    if np.array_equal(x, data):
            data = y
            y = None

    if type(data_object) is Cube:
        plot_defn = iplt._get_plot_defn(data_object, iris.coords.POINT_MODE, ndims = no_of_dims)
        if plot_defn.transpose:
            data = data.T
            x = x.T
            y = y.T

        if no_of_dims == 2:
            try:
                data, x = addcyclic(data, x)
                x, y = np.meshgrid(x, y)
            except:
                data, y = addcyclic(data, y)
                y, x = np.meshgrid(y, x)

    logging.debug("Shape of x: " + str(x.shape))
    if y is not None: logging.debug("Shape of y: " + str(y.shape))
    logging.debug("Shape of data: " + str(data.shape))

    return { "data": data, "x" : x, "y" : y }

def copy_attributes(source, dest):
    '''
     Copy all attributes from one object to another
    @param source: Object to copy attributes from
    @param dest: Object to copy attributes to
    @return: None
    '''
    if source:
        if isinstance(source, dict):
            dest.__dict__.update(source)
        else:
            dest.__dict__.update(source.__dict__)


def add_file_prefix(prefix, filepath):
    '''
        Add a prefix to a filename taking into account any path that might be present before that actual filename
    @param prefix: A string to prefix the filename with
    @param filepath: Filename, optionally including path
    @return: A string with the full path to the prefixed file
    '''
    import os.path
    filename = os.path.basename(filepath)
    path = os.path.dirname(filepath)
    return os.path.join(path,prefix+filename)


def parse_key_val_string(arguments, seperator):
    '''
        Takes a (comma) seperated list of keyword value pairs (seperated by =) and returns a dictionary with those keys and values
    @param arguments: A string which is a seperated list of keyword value pairs
    @param seperator: String which is used to split the string into a list
    @return: A dictionary of the keywords and values
    '''
    input_list = arguments.split(seperator)
    return parse_key_val_list(input_list)


def parse_key_val_list(input_list):
    '''
     Takes list of keyword value strings (seperated by =) and returns a dictionary with those keys and values
        NOTE - if a key has no value, the key is stored and given the value True
    @param input_list: A list of strings which are keyword value pairs seperated by =
    @return: A dictionary of the keywords and values
    '''
    key_val_dict = {}
    for element in input_list:
        key_value = element.split('=')
        key = key_value[0]
        value = key_value[1] if len(key_value) > 1 else True
        key_val_dict[key] = value

    return key_val_dict


def haversine(lat, lon, lat2, lon2):
    '''
        Computes the Haversine distance between two points
    '''
    import math
    R_E = 6378 # Radius of the earth in km
    lat1 = math.radians(lat)
    lat2 = math.radians(lat2)
    lon1 = math.radians(lon)
    lon2 = math.radians(lon2)
    arclen = 2*math.asin(math.sqrt((math.sin((lat2-lat1)/2))**2 + math.cos(lat1) * math.cos(lat2) * (math.sin((lon2-lon1)/2))**2))
    return arclen*R_E

class OrderedSet(collections.MutableSet):
    '''
    From http://code.activestate.com/recipes/576694/
    '''

    def __init__(self, iterable=None):
        self.end = end = []
        end += [None, end, end]         # sentinel node for doubly linked list
        self.map = {}                   # key --> [key, prev, next]
        if iterable is not None:
            self |= iterable

    def __len__(self):
        return len(self.map)

    def __contains__(self, key):
        return key in self.map

    def add(self, key):
        if key not in self.map:
            end = self.end
            curr = end[1]
            curr[2] = end[1] = self.map[key] = [key, curr, end]

    def discard(self, key):
        if key in self.map:
            key, prev, next = self.map.pop(key)
            prev[2] = next
            next[1] = prev

    def __iter__(self):
        end = self.end
        curr = end[2]
        while curr is not end:
            yield curr[0]
            curr = curr[2]

    def __reversed__(self):
        end = self.end
        curr = end[1]
        while curr is not end:
            yield curr[0]
            curr = curr[1]

    def pop(self, last=True):
        if not self:
            raise KeyError('set is empty')
        key = self.end[1][0] if last else self.end[2][0]
        self.discard(key)
        return key

    def __repr__(self):
        if not self:
            return '%s()' % (self.__class__.__name__,)
        return '%s(%r)' % (self.__class__.__name__, list(self))

    def __eq__(self, other):
        if isinstance(other, OrderedSet):
            return len(self) == len(other) and list(self) == list(other)
        return set(self) == set(other)

def apply_intersection_mask_to_two_arrays(array1, array2):
    '''
        Ensure two (optionally) masked arrays have the same mask.
        If both arrays are masked the intersection of the masks is used.
        If one array is masked and the other is not, the mask from the masked array is applied
          to the unmasked array
        If neither array is masked then both arrays are returned as masked arrays with an empty mask
    @param array1: An (optionally masked) array
    @param array2: Another (optionally masked) array
    @return: Two masked arrays with a common mask
    '''
    import numpy.ma as ma
    if isinstance(array1, ma.MaskedArray):
        if isinstance(array2, ma.MaskedArray):
            intersection_mask = ma.mask_or(array1.mask, array2.mask)
        else:
            intersection_mask = array1.mask
    else:
        if isinstance(array2, ma.MaskedArray):
            intersection_mask = array2.mask
        else:
            intersection_mask = False

    array1 = ma.array(array1, mask = intersection_mask)
    array2 = ma.array(array2, mask = intersection_mask)

    return array1, array2



from plotting.plot import Plotter
from data_io.Coord import CoordList, Coord
from data_io.ungridded_data import UngriddedData, Metadata
from numpy import array, arange

x = Coord(arange(10), Metadata('latitude'),'x')
coords = CoordList([x])

data1 = array([1,1.5,1,1.5,
              2,2,2.5,
              3,3.5,
              4,
              5,5.5,
              6,6.5,6,
              7,7,7.5,7,
              8,8.5,8,8,8,
              9,9,9.5,9,
              10,10.5,10])

data2 = array([1.5,1,1.5,1,1,
               2,2.5,2,
               3.5,
               5.5,5,5,
               6.5,6,
               7,7.5,7,7,7,
               8.5,8,8,8,8,8,8,
               9,9.5,9,
               10.5,10])


data_object1 = UngriddedData(data1, Metadata(standard_name='rain', long_name="TOTAL RAINFALL RATE: LS+CONV KG/M2/S", units="kg m-2 s-1", missing_value=-999), coords)
data_object2 = UngriddedData(data2, Metadata(standard_name='snow', long_name="TOTAL SNOWFALL RATE: LS+CONV KG/M2/S", units="kg m-2 s-1", missing_value=-999), coords)

Plotter([data_object1, data_object2],
        cmap = "RdBu",
        ylabel = "Overidden y",
        title = "Overidden Title",
        xrange = {"xmin" : 4, "xmax" : 8},
        yrange = {"ystep" : 0.5},
        valrange = {"vmin" : 1.2, "vstep" : 0.4},
        datagroups = [{"itemstyle" : None, "color" : None, "label" : None},
                      {"itemstyle" : None, "color" : None, "label" : None}],
        plot_type="histogram3d")




"""Tests for utility methods in subset module
"""
from nose.tools import istest, raises

from jasmin_cis.subsetting.subset import Subset

# Tests for _fix_longitude_limits
@istest
def test01_for_0_to_360():
    result = Subset._fix_longitude_limits(0, 90, 0, 360)
    assert (result == (0, 90))


@istest
def test02_for_0_to_360():
    result = Subset._fix_longitude_limits(10, 350, 0, 360)
    assert (result == (10, 350))


@istest
def test03_for_0_to_360():
    result = Subset._fix_longitude_limits(0, 359.9, 0, 360)
    assert (result == (0, 359.9))


@istest
def test04_for_0_to_360():
    result = Subset._fix_longitude_limits(180, 359.9, 0, 360)
    assert (result == (180, 359.9))


@istest
def test05_for_0_to_360():
    result = Subset._fix_longitude_limits(0, 360, 0, 360)
    assert (result == (0, 360))


@istest
def test06_for_0_to_360():
    result = Subset._fix_longitude_limits(0, 0, 0, 360)
    assert (result == (0, 0))


@istest
def test07_for_0_to_360():
    result = Subset._fix_longitude_limits(360, 360, 0, 360)
    assert (result == (360, 360))


@istest
def test08_for_0_to_360():
    result = Subset._fix_longitude_limits(0.1, -0.1, 0, 360)
    assert (result == (0.1, 359.9))


@istest
def test09_for_0_to_360():
    result = Subset._fix_longitude_limits(-1, 359, 0, 360)
    assert (result == (359, 359))


@istest
def test10_for_0_to_360():
    result = Subset._fix_longitude_limits(-1, 359, 0, 360)
    assert (result == (359, 359))


@istest
def test01_for_minus180_to_180():
    result = Subset._fix_longitude_limits(0, 0, -180, 180)
    assert (result == (0, 0))


@istest
def test02_for_minus180_to_180():
    result = Subset._fix_longitude_limits(-180, -180, -180, 180)
    assert (result == (-180, -180))


@istest
def test03_for_minus180_to_180():
    result = Subset._fix_longitude_limits(180, 180, -180, 180)
    assert (result == (180, 180))


@istest
def test04_for_minus180_to_180():
    result = Subset._fix_longitude_limits(-180, 180, -180, 180)
    assert (result == (-180, 180))


@istest
def test05_for_minus180_to_180():
    result = Subset._fix_longitude_limits(1, -180.1, -180, 180)
    assert (result == (1, 179.9))


@istest
def test06_for_minus180_to_180():
    result = Subset._fix_longitude_limits(181, 179, -180, 180)
    assert (result == (-179, 179))


# Ambiguous as to whether 0,360 or -180,180
@istest
def test01_for_0_to_180():
    result = Subset._fix_longitude_limits(0, 180, 0, 180)
    assert (result == (0, 180))


@istest
def test02_for_0_to_180():
    result = Subset._fix_longitude_limits(0, 360, 0, 180)
    assert (result == (0, 0))


@istest
def test01_for_minus1_to_180():
    result = Subset._fix_longitude_limits(181, 179, -1, 180)
    assert (result == (-179, 179))


if __name__ == '__main__':
    import nose
    nose.runmodule()



from glob import glob
from pyhdf.error import HDF4Error
import numpy as np
import hdf

def dictkeysappend(dict1,dict2,axis=-1):
    '''
    Appends data in dict2 to already existing keys in a dictionary dict1.
    If that key does not exist, it is created
    '''
    for key in dict2.keys():
        try:
            dict1[key] = np.concatenate((dict1[key],dict2[key]),axis=axis)
        except KeyError:
            dict1[key] = dict2[key]
    return dict1

def readin_cloudsat_cwc_rvod(filenames,sds,outdata=None):
    '''
    Reads in data from CloudSat 2B-CWC-RVOD files.
    Also reads in some geolocation data (lat,lon,TAI time) into the output dictionary.
    Setting outdata allows the data to be read into an already exisiting dictionary
    '''
    names = sds+['Latitude','Longitude','TAI_start','Profile_time']
    if outdata == None:
        outdata = {}
    for file in filenames:
        try:
            print '...Reading file ' + file
            data = hdf.read_hdf4(file,names,vdata=True)
            data['Profile_time'] = data['Profile_time'] + data['TAI_start']
            outdata = dictkeysappend(outdata,data,axis=0)
        except HDF4Error:
            print HDF4Error, ' in readin_cloudsat_cwc_rvod', file
    return outdata


def plot(data):

    import matplotlib.pyplot as plt

    plt.imshow(data['RVOD_liq_water_content'][3000:3600,:103].transpose(),vmin=0)
    plt.yticks([0,25,50,75,100],[25,18.75,12.5,6.25,0])
    plt.ylabel('~Height above MSL (km)')
    plt.title('CloudSat RVOD Liquid Water Content')
    cm = plt.colorbar()
    cm.set_label('mg m$^{-3}$')
    plt.show()

if __name__ == '__main__':

    import numpy as np

    # get a list of files to read
    folder = "/home/david/Data"
    #filenames = glob(folder + "/*" + "CS_2B-CWC-RVOD_GRANULE_P_R04_E02.hdf")
    filenames = [folder + "/2007180012247_06214_CS_2B-CWC-RVOD_GRANULE_P_R04_E02.hdf"]
    filenames = np.sort(filenames)

    # read the data
    data = readin_cloudsat_cwc_rvod(filenames,['RVOD_liq_water_content','Height','RVOD_ice_water_content'])

    # print data
    print "\nkeys are: " , data.keys()

    for key, value in data.iteritems():
        print "\n" + key + " :", value, len(value), np.nanmin(value), np.nanmax(value)

    # plot data
    plot(data)



from jasmin_cis.data_io.products.AProduct import __get_class
from nose.tools import istest, eq_
from jasmin_cis.test.test_files.data import *
from jasmin_cis.parse import parse_args

@istest
def can_overide_default_product():
    from jasmin_cis.data_io.products.products import Caliop_L2, Xglnwa
    filename = valid_caliop_l2_filename
    eq_(__get_class(filename), Caliop_L2)
    eq_(__get_class(filename, "Xglnwa"), Xglnwa)

@istest
def should_raise_error_with_unknown_product_specified():
    filename = netcdf_file_with_incorrect_file_extension
    try:
        parse_args(["plot", valid_variable_in_valid_filename + ":" + filename + "::::unknownproduct"])
        assert False
    except SystemExit as e:
        if e.code != 2:
            raise e




from jasmin_cis.data_io.read import read_data
from jasmin_cis.plotting.plot import Plotter
from copy import deepcopy

variable = "Effective_Optical_Depth_Average_Ocean"
filename = "/home/daniel/Code/jasmin_cis/jasmin_cis/test/test_files/MOD04_L2.A2010001.2255.005.2010005215814.hdf"
#wavelength_index = 3

data_object = read_data([filename], variable)

# Do all coords have the same dimensionality?
coord_shape = None
for coord in data_object.coords():
    if coord_shape == None:
        coord_shape = coord.shape
    else:
        if coord_shape != coord.shape:
            raise Exception("All coords must be of the same shape")

print "All coords are of the shape: " + str(coord_shape)

print "Shape of data: " + str(data_object.shape)

if len(data_object.shape) > len(coord_shape):
    print "Shape of data has " + str(len(data_object.shape) - len(coord_shape)) + " more dimension(s) than its coordinates"

for wavelength_index in range(0,7,1):
    print wavelength_index
    sliced_data_object = deepcopy(data_object)
    sliced_data_object.data = data_object.data[wavelength_index,:,:]
    sliced_data_object.shape = [203, 135]

    Plotter([sliced_data_object], out_filename="/home/daniel/Desktop/modis" + str(wavelength_index))


"""
Custom CIS exceptions
"""

class CISError(Exception):
    pass

class InvalidPlotTypeError(CISError):
    pass

class InvalidDimensionError(CISError):
    pass

class InvalidVariableError(CISError):
    pass

class InconsistentDimensionsError(CISError):
    pass

class InvalidPlotFormatError(CISError):
    pass

class InvalidFileExtensionError(CISError):
    pass

class InvalidDataTypeError(CISError):
    pass

class InvalidColocationMethodError(CISError):
    pass

class InvalidLineStyleError(CISError):
    pass

class InvalidHistogramStyleError(CISError):
    pass

class CoordinateNotFoundError(CISError):
    pass

class DuplicateCoordinateError(CISError):
    pass

class ClassNotFoundError(CISError):
    pass

class InvalidCommandLineOptionError(CISError):
    pass

class InvalidNumberOfDatagroupsSpecifiedError(CISError):
    pass

class NotEnoughAxesSpecifiedError(CISError):
    pass

class InvalidSliceIndexError(CISError):
    pass

class NoDataInSubsetError(CISError):
    pass



from nose.tools import istest, raises
import numpy
from jasmin_cis.data_io.Coord import Coord, CoordList
from jasmin_cis.data_io.ungridded_data import Metadata
from jasmin_cis.exceptions import DuplicateCoordinateError

def create_dummy_coordinates_list():
    coord1 = Coord(numpy.array([5,4]),Metadata(standard_name='testY'),axis='Y')
    coord2 = Coord(numpy.array([5,4]),Metadata(standard_name='testX'),axis='X')
    return CoordList([coord1,coord2])

@istest
def can_create_a_valid_list_of_coordinates():
    list = create_dummy_coordinates_list()
    assert(len(list)==2)
    assert(list[0].standard_name=='testY')
    assert(list[1].standard_name=='testX')
    assert(list[0].axis=='Y')
    assert(list[1].axis=='X')

@istest
def can_append_to_list_of_coordinates():
    list = create_dummy_coordinates_list()
    list.append(Coord(numpy.array([5,4]),Metadata(standard_name='testZ'),axis='Z'))
    assert(len(list)==3)
    assert(list[2].standard_name=='testZ')
    assert(list[2].axis=='Z')

@istest
@raises(DuplicateCoordinateError)
def append_a_duplicate_to_a_list_of_coordinates_fails():
    list = create_dummy_coordinates_list()
    list.append(Coord(numpy.array([5,4]),Metadata(standard_name='testX'),axis='X'))

@istest
def can_find_a_coord_from_a_list_of_coordinates():
    list = create_dummy_coordinates_list()
    coord = list.get_coord(standard_name='testX')
    assert(coord.standard_name=='testX')
    assert(coord.axis=='X')

@istest
def can_find_many_coords_from_a_list_of_coordinates():
    list = create_dummy_coordinates_list()
    list.append(Coord(numpy.array([5,4]),Metadata(name='testZ'),axis='Z'))
    list.append(Coord(numpy.array([5,4]),Metadata(name='testZ',standard_name='my_other_z_coord')))
    assert(len(list)==4)
    coords = list.get_coords(name='testZ')
    assert(len(coords)==2)
    assert(coords[0].name()=='testZ')
    assert(coords[0].axis=='Z')
    assert(coords[1].axis=='')



'''
Module for reading data. This is just an abstract layer for making the choice between gridded and ungridded - the actual reading is done elsewhere
'''

def read_data(filenames, variable, product=None):
    '''
    Read a specific variable from a list of files
    Files can be either gridded or ungridded but not a mix of both.
    First tries to read data as gridded, if that fails, tries as ungridded.
    
    @param filenames:   The filenames of the files to read
    @param variable:    The variable to read from the files
        
    @return:  The specified data with unnecessary dimensions removed
    '''
    from jasmin_cis.data_io.products.AProduct import get_data

    # if filenames is not a list, make it a list of 1 element
    if not isinstance(filenames, list): filenames = [ filenames ]

    return get_data(filenames, variable, product)


def read_coordinates(filenames, product=None):
    '''
    Read the coordinates from a file

    @param filenames:   The filename of the files to read
        
    @return: A CoordList object
    '''
    from jasmin_cis.data_io.products.AProduct import get_coordinates

    # if filenames is not a list, make it a list of 1 element
    if not isinstance(filenames, list): filenames = [ filenames ]

    return get_coordinates(filenames, product)




"""
Module containing hdf file functions
"""

from pyhdf import SD
import numpy as np
from pyhdf.HDF import *
from pyhdf.VS import *

def read_hdf4(filename, names=None, calipso_scaling=None, vdata=None,datadict=None):
    """
    Reads data from a HDF4 file into a dictionary. Automatically applies the
    scaling factors and offsets to the data arrays often found in NASA HDF-EOS
    data.
    
    Returns:
        A dictionary containing data for requested variables.
        Missing data is replaced by NaN.
        
    Arguments:
        filename    -- The name (with path) of the HDF file to read.
        names       -- A sequence of variable (dataset) names to read from the
                       file (default None, causing all variables to be read).
                       The names must appear exactly as in in the HDF file.

        calipso_scaling -- If set, the code will apply offsets and scaling in
                       the order used by CALIPSO data (rather than the MODIS
                       method). The two methods are:
                       MODIS:   (data - offset) * scale_factor
                       CALIPSO: data/scalefactor + offset.
        cloudsatPRECIP -- If set, reads in cloudsat data from the vdata style 
                       hdf file. For 2C-PRECIP-COLUMN.  Reads in dictionary of 
                       numpy arrays
    """

    if not vdata:
    # MODIS/Calipso data
    # Open the file.
        datafile = SD.SD(filename)
        # List of required variable names.
        if not names:
            names = datafile.datasets()
            # Create dictionary to hold data arrays for returning.
        if datadict==None:
            datadict = {}
            # Get data.
        for name in names:
            sds = datafile.select(name) # SDS object.
            data = sds.get()    # Extract raw data.
            attributes = sds.attributes()
            fillvalue = attributes.get('_FillValue', None)  # Missing data.
            wfill_mask = np.where(data == fillvalue, np.nan, 1)
            data = data * wfill_mask
            offset  = attributes.get('add_offset', 0)   # Offsets and scaling.
            scalefactor = attributes.get('scale_factor', 1)
            if not calipso_scaling:
                data = (data - offset) * scalefactor    # MODIS.
            else:
                data = (data/scalefactor) + offset      # Calipso.
            datadict[name] = data

    else:
        # CloudSatPRECIP data
        # Open file
        datafile = HDF(filename)
        vs =  datafile.vstart()
        # List of required variable names
        if not names:
            names = vs.vdatainfo()
            names = zip(*names)
            names = names[0]
        if datadict == None:
            datadict = {}
        failednames = []
        for name in names:
            try:
                vd = vs.attach(name)
                data = vd.read(nRec = vd.inquire()[0])
                for x in range(0,len(data)):
                    data[x] = data[x][0]
                data = np.array(data)
                try: #Deal with missing data
                    missingval = vd.attrinfo()['missing'][2]
                    wfillmask = 0
                    wfillmask = np.where(data == missingval, np.nan, 1)
                    data = data * wfillmask
                except KeyError:
                    pass
                datadict[name] = data
                vd.detach()
            except:
                failednames.append(name)
        vs.end()
        datafile.close()
        if len(failednames) > 0:
            datadict = read_hdf4(filename,failednames,datadict=datadict)

    return datadict

def read_hdf4_metadata(filename,name):
    '''Retrieves long name and units from hdf4 file, for a single variable.'''
    datafile = SD.SD(filename)
    sds =datafile.select(name)
    attributes = sds.attributes()
    longName = attributes.get('long_name')
    units = attributes.get('units')
    validRange = attributes.get('valid_range')
    return longName,units,validRange




